import warnings

import numpy as np
from scipy.interpolate import RegularGridInterpolator as RGI

from .methods import bin_width_doane_div5, get_bad_vals, methods
from .contours import find_contours_level, get_quantile_levels


class ContourSpacingTooLarge(UserWarning):
    pass


class KernelDensityEstimator:
    def __init__(self, rtdc_ds):
        self.rtdc_ds = rtdc_ds

    @staticmethod
    def apply_scale(a, scale, feat):
        """Helper function for transforming an aray to log-scale

        Parameters
        ----------
        a: np.ndarray
            Input array
        scale: str
            If set to "log", take the logarithm of `a`; if set to
            "linear" return `a` unchanged.
        feat: str
            Feature name (required for debugging)

        Returns
        -------
        b: np.ndarray
            The scaled array

        Notes
        -----
        If the scale is not "linear", then a new array is returned.
        All warnings are suppressed when computing `np.log(a)`, as
        `a` may have negative or nan values.
        """
        if scale == "linear":
            b = a
        elif scale == "log":
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                b = np.log(a)
                if len(w):
                    # Tell the user that the log-transformation issued
                    # a warning.
                    warnings.warn(f"Invalid values encounterd in np.log "
                                  f"while scaling feature '{feat}'!")
        else:
            raise ValueError(f"`scale` must be either 'linear' or 'log', "
                             f"got '{scale}'!")
        return b

    @staticmethod
    def get_spacing(a, method, scale="linear", method_kw=None,
                    feat="undefined", ret_scaled=False):
        """Convenience function for computing the contour spacing

        Parameters
        ----------
        a: ndarray
            feature data
        scale: str
            how the data should be scaled ("log" or "linear")
        method: callable
            KDE spacing method to use
        method_kw: dict
            keyword arguments to `method`
        feat: str
            feature name for debugging
        ret_scaled: bool
            whether to return the scaled array of `a`
        """
        if method_kw is None:
            method_kw = {}
        # Apply scale (no change for linear scale)
        asc = KernelDensityEstimator.apply_scale(a, scale, feat)
        # Apply multiplicator
        acc = method(asc, **method_kw)
        if ret_scaled:
            return acc, asc
        else:
            return acc

    def get_contour(self, xax="area_um", yax="deform", xacc=None, yacc=None,
                    kde_type="histogram", kde_kwargs=None, xscale="linear",
                    yscale="linear"):
        """Evaluate the kernel density estimate for contour plots

        Parameters
        ----------
        xax: str
            Identifier for X axis (e.g. "area_um", "aspect", "deform")
        yax: str
            Identifier for Y axis
        xacc: float
            Contour accuracy in x direction
        yacc: float
            Contour accuracy in y direction
        kde_type: str
            The KDE method to use
        kde_kwargs: dict
            Additional keyword arguments to the KDE method
        xscale: str
            If set to "log", take the logarithm of the x-values before
            computing the KDE. This is useful when data are
            displayed on a log-scale. Defaults to "linear".
        yscale: str
            See `xscale`.

        Returns
        -------
        X, Y, Z : coordinates
            The kernel density Z evaluated on a rectangular grid (X,Y).
        """
        warnings.warn("`get_contour` is deprecated; please use "
                      "`get_raster` instead", DeprecationWarning)
        return self.get_raster(
            xax=xax, yax=yax, xacc=xacc, yacc=yacc,
            kde_type=kde_type, kde_kwargs=kde_kwargs,
            xscale=xscale, yscale=yscale
        )

    def get_contour_lines(self, quantiles=None, xax="area_um", yax="deform",
                          xacc=None, yacc=None, kde_type="histogram",
                          kde_kwargs=None, xscale="linear", yscale="linear",
                          ret_levels=False):
        """Compute contour lines for a given kernel kensity estimate.

        Parameters
        ----------
        quantiles: list or array of floats
            KDE Quantiles for which contour levels are computed. The
            values must be between 0 and 1. If set to None, use
            [0.5, 0.95] as default.
        xax: str
            Identifier for X axis (e.g. "area_um", "aspect", "deform")
        yax: str
            Identifier for Y axis
        xacc: float
            Contour accuracy in x direction
            if set to None, will use :func:`bin_width_doane_div5`
        yacc: float
            Contour accuracy in y direction
            if set to None, will use :func:`bin_width_doane_div5`
        kde_type: str
            The KDE method to use
        kde_kwargs: dict
            Additional keyword arguments to the KDE method
        xscale: str
            If set to "log", take the logarithm of the x-values before
            computing the KDE. This is useful when data are
            displayed on a log-scale. Defaults to "linear".
        yscale: str
            See `xscale`
        ret_levels: bool
            If set to True, return the levels of the contours
            (default: False)

        Returns
        -------
        contour_lines: list of lists (of lists)
            For every number in `quantiles`, this list contains a list of
            corresponding contour lines. Each contour line is a 2D
            array of shape (N, 2), where N is the number of points in the
            contour line.
        levels: list of floats
            The density levels corresponding to each number in `quantiles`.
            Only returned if `ret_levels` is set to True.
        """
        if not quantiles:
            quantiles = [0.5, 0.95]
        try:
            x, y, density = self.get_raster(
                xax=xax,
                yax=yax,
                xacc=xacc,
                yacc=yacc,
                xscale=xscale,
                yscale=yscale,
                kde_type=kde_type,
                kde_kwargs=kde_kwargs,
            )
        except ValueError:
            # most-likely there is nothing to compute a contour for
            return []
        if density.shape[0] < 3 or density.shape[1] < 3:
            warnings.warn("Contour not possible; spacing may be too large!",
                          ContourSpacingTooLarge)
            return []
        levels = get_quantile_levels(
            density=density,
            x=x,
            y=y,
            xp=self.rtdc_ds[xax][self.rtdc_ds.filter.all],
            yp=self.rtdc_ds[yax][self.rtdc_ds.filter.all],
            q=np.array(quantiles),
            normalize=False)
        contours = []
        # Normalize levels to [0, 1]
        nlevels = np.array(levels) / density.max()
        for nlev in nlevels:
            # make sure that the contour levels are not at the boundaries
            if not (np.allclose(nlev, 0, atol=1e-12, rtol=0)
                    or np.allclose(nlev, 1, atol=1e-12, rtol=0)):
                cc = find_contours_level(
                    density, x=x, y=y, level=nlev)
                contours.append(cc)
            else:
                contours.append([])
        if ret_levels:
            return contours, levels
        else:
            return contours

    def get_raster(self, xax="area_um", yax="deform", xacc=None, yacc=None,
                   kde_type="histogram", kde_kwargs=None, xscale="linear",
                   yscale="linear"):
        """Evaluate the kernel density estimate on a grid

        Parameters
        ----------
        xax: str
            Identifier for X axis (e.g. "area_um", "aspect", "deform")
        yax: str
            Identifier for Y axis
        xacc: float
            Contour accuracy in x direction
            if set to None, will use :func:`bin_width_doane_div5`
        yacc: float
            Contour accuracy in y direction
            if set to None, will use :func:`bin_width_doane_div5`
        kde_type: str
            The KDE method to use
        kde_kwargs: dict
            Additional keyword arguments to the KDE method
        xscale: str
            If set to "log", take the logarithm of the x-values before
            computing the KDE. This is useful when data are
            displayed on a log-scale. Defaults to "linear".
        yscale: str
            See `xscale`.

        Returns
        -------
        X, Y, Z : coordinates
            The kernel density Z evaluated on a rectangular grid (X,Y).
        """
        if kde_kwargs is None:
            kde_kwargs = {}
        xax = xax.lower()
        yax = yax.lower()
        kde_type = kde_type.lower()
        if kde_type not in methods:
            raise ValueError(f"Not a valid kde type: {kde_type}!")

        # Get data
        x = self.rtdc_ds[xax][self.rtdc_ds.filter.all]
        y = self.rtdc_ds[yax][self.rtdc_ds.filter.all]

        xacc_sc, xs = self.get_spacing(
            a=x,
            feat=xax,
            scale=xscale,
            method=bin_width_doane_div5,
            ret_scaled=True)

        yacc_sc, ys = self.get_spacing(
            a=y,
            feat=yax,
            scale=yscale,
            method=bin_width_doane_div5,
            ret_scaled=True)

        if xacc is None or xacc == 0:
            xacc = xacc_sc

        if yacc is None or yacc == 0:
            yacc = yacc_sc

        # Ignore infs and nans
        bad = get_bad_vals(xs, ys)
        xc = xs[~bad]
        yc = ys[~bad]

        if xc.size:
            # Compute the mesh for rastering
            xnum = int(np.ceil((xc.max() - xc.min()) / xacc))
            ynum = int(np.ceil((yc.max() - yc.min()) / yacc))

            xlin = np.linspace(xc.min(), xc.max(), xnum, endpoint=True)
            ylin = np.linspace(yc.min(), yc.max(), ynum, endpoint=True)

            xmesh, ymesh = np.meshgrid(xlin, ylin, indexing="ij")

            # Compute the KDE for each point on the mesh
            kde_fct = methods[kde_type]
            density = kde_fct(events_x=xs, events_y=ys,
                              xout=xmesh, yout=ymesh,
                              **kde_kwargs)
        else:
            xmesh, ymesh = np.meshgrid([0, 1], [0, 1], indexing="ij")
            density = np.array(np.nan * xmesh)

        # Convert mesh back to linear scale if applicable
        if xscale == "log":
            xmesh = np.exp(xmesh)
        if yscale == "log":
            ymesh = np.exp(ymesh)

        return xmesh, ymesh, density

    def get_scatter(self, xax="area_um", yax="deform", positions=None,
                    kde_type="histogram", kde_kwargs=None, xscale="linear",
                    yscale="linear"):
        """Evaluate the kernel density estimate for scatter plots

        The KDE is evaluated with the `kde_type` function for every point.

        Parameters
        ----------
        xax: str
            Identifier for X axis (e.g. "area_um", "aspect", "deform")
        yax: str
            Identifier for Y axis
        positions: list of two 1d ndarrays or ndarray of shape (2, N)
            The positions where the KDE will be computed. Note that
            the KDE estimate is computed from the points that
            are set in `self.rtdc_ds.filter.all`.
        kde_type: str
            The KDE method to use, see :const:`.kde_methods.methods`
        kde_kwargs: dict
            Additional keyword arguments to the KDE method
        xscale: str
            If set to "log", take the logarithm of the x-values before
            computing the KDE. This is useful when data are
            displayed on a log-scale. Defaults to "linear".
        yscale: str
            See `xscale`.

        Returns
        -------
        density : 1d ndarray
            The kernel density evaluated for the filtered data points.
        """
        if kde_kwargs is None:
            kde_kwargs = {}
        xax = xax.lower()
        yax = yax.lower()
        kde_type = kde_type.lower()
        if kde_type not in methods:
            raise ValueError(f"Not a valid kde type: {kde_type}!")

        # Get data
        x = self.rtdc_ds[xax][self.rtdc_ds.filter.all]
        y = self.rtdc_ds[yax][self.rtdc_ds.filter.all]

        # Apply scale (no change for linear scale)
        xs = self.apply_scale(x, xscale, xax)
        ys = self.apply_scale(y, yscale, yax)

        if positions is None:
            posx = None
            posy = None
        else:
            posx = self.apply_scale(positions[0], xscale, xax)
            posy = self.apply_scale(positions[1], yscale, yax)

        kde_fct = methods[kde_type]
        if len(x):
            density = kde_fct(events_x=xs, events_y=ys,
                              xout=posx, yout=posy,
                              **kde_kwargs)
        else:
            density = np.array([])

        return density

    def get_at(self, xax="area_um", yax="deform", positions=None,
               kde_type="histogram", kde_kwargs=None, xscale="linear",
               yscale="linear"):
        """Evaluate the kernel density estimate for specific events

        The KDE is computed via linear interpolation from the output
        of `get_raster`.

        Parameters
        ----------
        xax: str
            Identifier for X axis (e.g. "area_um", "aspect", "deform")
        yax: str
            Identifier for Y axis
        positions: list of two 1d ndarrays or ndarray of shape (2, N)
            The positions where the KDE will be computed. Note that
            the KDE estimate is computed from the points that
            are set in `self.rtdc_ds.filter.all`.
        kde_type: str
            The KDE method to use, see :const:`.kde_methods.methods`
        kde_kwargs: dict
            Additional keyword arguments to the KDE method
        xscale: str
            If set to "log", take the logarithm of the x-values before
            computing the KDE. This is useful when data are
            displayed on a log-scale. Defaults to "linear".
        yscale: str
            See `xscale`.

        Returns
        -------
        density : 1d ndarray
            The kernel density evaluated for the filtered events.
        """
        if kde_kwargs is None:
            kde_kwargs = {}
        xax = xax.lower()
        yax = yax.lower()
        kde_type = kde_type.lower()
        if kde_type not in methods:
            raise ValueError(f"Not a valid kde type: {kde_type}!")

        # Get data
        x = self.rtdc_ds[xax][self.rtdc_ds.filter.all]
        y = self.rtdc_ds[yax][self.rtdc_ds.filter.all]

        # Apply scale (no change for linear scale)
        xs = self.apply_scale(x, xscale, xax)
        ys = self.apply_scale(y, yscale, yax)

        if positions:
            xs = self.apply_scale(positions[0], xscale, xax)
            ys = self.apply_scale(positions[1], yscale, yax)

        if len(x):
            xr, yr, density_grid = self.get_raster(xax=xax,
                                                   yax=yax,
                                                   kde_type=kde_type,
                                                   kde_kwargs=kde_kwargs,
                                                   xscale=xscale,
                                                   yscale=yscale)

            # Apply scale (no change for linear scale)
            xrs = self.apply_scale(xr, xscale, xax)
            yrs = self.apply_scale(yr, yscale, yax)

            # 'scipy.interp2d' has been removed in SciPy 1.14.0
            # https://scipy.github.io/devdocs/tutorial/interpolate/interp_transition_guide.html
            interp_func = RGI((xrs[:, 0], yrs[0, :]),
                              density_grid,
                              method="linear",
                              bounds_error=False,
                              fill_value=np.nan)
            density = interp_func((xs, ys))

        else:
            density = np.array([])

        return density



"""vid2dc

Convert video files to .rtdc files. This script generates a valid .rtdc
file from a video file (e.g. an .avi file). This script does not perform
segmentation and feature extraction, a feature that is available via the
ChipStream package (https://github.com/DC-analysis/ChipStream).

Prerequisites:

    pip install dclab[tdms] imageio[pyav] click

Author: Paul Müller
License: GPLv3+
"""
import configparser
import logging
import pathlib
import time
import traceback

import click
from dclab.rtdc_dataset import fmt_tdms
from dclab.util import hashfile, hashobj
from dclab import RTDCWriter
import h5py
import imageio.v3 as iio
import numpy as np


logger = logging.getLogger("vid2dc")
logging.basicConfig(level=logging.INFO)


__version__ = "2025.07.07"


@click.command()
@click.argument("path_in",
                type=click.Path(exists=True,
                                dir_okay=True,
                                resolve_path=True,
                                path_type=pathlib.Path))
@click.argument("path-out",
                required=False,
                type=click.Path(dir_okay=False,
                                writable=True,
                                resolve_path=True,
                                path_type=pathlib.Path),
                )
def vid2dc(path_in: pathlib.Path,
           path_out: pathlib.Path | None = None):
    """Convert raw video data to raw .rtdc data

    Use this command to convert a video file (e.g. from a legacy
    .tdms measurement or from a custom setup) to a raw .rtdc file.
    No segmentation or feature extraction will take place.

    You may also specify an input directory instead of a single file
    for batch conversion.
    """
    if path_in.is_dir():
        if path_out is not None:
            raise NotImplementedError("If you pass an input directory, you "
                                      "must omit the output directory.")
        for ext in [".avi", ".mp4", ".mpg", ".wmv", ".mkv"]:
            for pp in path_in.rglob(f"*{ext}"):
                try:
                    convert_video_to_dc(video_path=pp)
                except BaseException:
                    elog = pp.with_name(pp.stem + "_vid2dc_error.log")
                    elog.write_text(traceback.format_exc())
                    click.secho(f"Could not process {pp}, wrote error log "
                                f"to {elog}.",
                                bold=True, fg="red")
    else:
        convert_video_to_dc(video_path=path_in, dc_path=path_out)


def convert_video_to_dc(video_path: str | pathlib.Path,
                        dc_path: str | pathlib.Path = None,
                        ):
    """Convert an .avi file to an .rtdc file

    A valid .rtdc file is produced. Metadata are guessed, but you
    can put metadata in camera.ini files. This is not fully supported yet.
    Please check the metadata of the output .rtdc file.

    Parameters
    ----------
    video_path:
        Path to a video file (e.g. an .avi file) to convert to .rtdc.
    dc_path:
        Path to an .rtdc file that is created from the video file.
        If the .rtdc file already exists, it will only be overridden
        if the input file or metadata changed.
    """
    video_path = pathlib.Path(video_path)
    if dc_path is None:
        dc_path = video_path.with_name(video_path.stem + "_raw.rtdc")
    dc_path = pathlib.Path(dc_path)

    # get the dataset configuration
    config, log_paths = fmt_tdms.RTDC_TDMS.extract_tdms_config(
        video_path, ret_source_files=True, ignore_missing=True)

    # RecoTeem stores an ini file next to the measurement
    prt = video_path.with_suffix(".ini")
    if prt.exists():
        log_paths.append(prt)

    # read logs
    logs = {}
    for ppl in log_paths:
        logs[ppl.name] = ppl.read_text().split("\n")

    vmeta = iio.improps(video_path)
    width = vmeta.shape[2]
    height = vmeta.shape[1]

    config["imaging"].setdefault("frame rate", 2000)
    config["imaging"].setdefault("pixel size", 0.34)
    config["imaging"]["roi size x"] = width
    config["imaging"]["roi size y"] = height

    # extract additional information from an optional ini file (RecoTeem)
    if prt.exists():
        # measurement index
        if prt.name.startswith("M") and prt.name.count("_"):
            mid = video_path.name.split("_")[0]
            config["experiment"]["run index"] = int(mid[1:])
        cp = configparser.RawConfigParser(allow_no_value=True)
        # Allow upper-case keys (comment for sample name)
        cp.optionxform = lambda x: x
        cp.read(prt)
        rt_fr = cp.getint("set parameter", "frame rate [fps]", fallback=None)
        # frame rate
        if rt_fr is not None:
            config["imaging"]["frame rate"] = rt_fr
        rt_dur = cp.getfloat("set parameter", "set record time [s]",
                             fallback=None)
        # duration
        if rt_dur is not None:
            config["online_filter"]["target duration"] = rt_dur / 60
        # sample name
        if cp.has_section("Optional Text"):
            descr = list(cp["Optional Text"].keys())[0]
            config["experiment"]["sample"] = descr

    # Set the time of the input file
    tse = video_path.stat().st_mtime
    loct = time.localtime(tse)
    # Start time of measurement ('HH:MM:SS')
    timestr = time.strftime("%H:%M:%S", loct)
    config["experiment"].setdefault("time", timestr)
    # Date of measurement ('YYYY-MM-DD')
    datestr = time.strftime("%Y-%m-%d", loct)
    config["experiment"].setdefault("date", datestr)
    config["experiment"].setdefault("sample", video_path.name)

    config["setup"].setdefault("chip region", "channel")
    config["setup"].setdefault("flow rate", 0)
    config["setup"].setdefault("software version", f"vid2dc {__version__}")

    config = dict(config)
    if "filtering" in config:
        config.pop("filtering")

    # Compute a unique hash of the input files
    video_hash = hashobj([hashobj(config),
                          hashfile(video_path),
                          video_path.stat().st_size,
                          ])

    # If the output file already exists, check whether the hashes match.
    log_name = "vid2dc"
    if dc_path.exists():
        with h5py.File(dc_path) as h5:
            if (log_name in h5.get("logs", [])
                    and h5["logs"][log_name].attrs.get("hash") == video_hash):
                # We can reuse this file
                logger.info(f"Reusing existing file {dc_path}")
                return
            else:
                logger.info(f"Rewriting existing {dc_path} (hash mismatch)")
                dc_path.unlink()

    vid = iio.imiter(video_path)

    # This is the actual workload of this function. Populate the .rtdc
    # file with the image data.
    logger.info(f"Writing .rtdc file {dc_path}")
    with RTDCWriter(dc_path, mode="reset") as hw:
        # store the image data to the output file in chunks
        chunk_size = hw.get_best_nd_chunks(item_shape=(height, width),
                                           item_dtype=np.uint8)[0]
        image_chunk = np.zeros((chunk_size, height, width), dtype=np.uint8)
        ii = 0
        for image in vid:
            if len(image.shape) == 3:
                image = image[:, :, 0]
            image_chunk[ii] = image
            ii += 1
            if ii == chunk_size:
                ii = 0
                hw.store_feature("image", image_chunk)
        # store the rest
        if ii:
            hw.store_feature("image", image_chunk[:ii])

        event_count = hw.h5file["events/image"].shape[0]
        config["experiment"]["event count"] = event_count
        hw.store_metadata(config)

        # store the input file information as a log
        hw.store_log(
            name=log_name,
            lines=[
                f"Input filename: {video_path.name}",
                f"Output filename: {dc_path.name}",
                f"Input file size: {video_path.stat().st_size}",
                f"Input last edited: {video_path.stat().st_mtime}",
                f"Unique conversion hash: {video_hash}",
            ]
        )

        # store time
        hw.store_feature(
            "time",
            np.arange(event_count) / config["imaging"]["frame rate"])
        # index starts at 1
        hw.store_feature("index", np.arange(1, event_count + 1))
        # frame is an ascending number, just start with 1
        hw.store_feature("frame", np.arange(1, event_count + 1))

        # only store hash attribute when successful
        hw.h5file["logs"][log_name].attrs["hash"] = video_hash


if __name__ == "__main__":
    vid2dc()



import h5py
import numpy as np
import pytest

import dclab

from helper_methods import retrieve_data


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_af_brightness_bc():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    ds = dclab.new_dataset(path)
    # sanity checks
    assert "bright_perc_10" not in ds.features_innate
    assert "bright_perc_90" not in ds.features_innate
    # ignore first event (no image data)
    for ii in range(1, len(ds)):
        image_corr = np.array(ds["image"][ii], dtype=int) - ds["image_bg"][ii]
        mask = ds["mask"][ii]
        p10 = np.percentile(image_corr[mask], 10)
        p90 = np.percentile(image_corr[mask], 90)
        assert np.max(np.abs(image_corr)) < 50, "sanity check"
        assert ds["bright_perc_10"][ii] == p10
        assert ds["bright_perc_90"][ii] == p90


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_af_brightness_bc_with_offset():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    with h5py.File(path, "a") as h5:
        bg_off = np.linspace(0, 1, len(h5["events/deform"]))
        h5["events/bg_off"] = bg_off

    ds = dclab.new_dataset(path)
    # sanity checks
    assert "bright_perc_10" not in ds.features_innate
    assert "bright_perc_90" not in ds.features_innate
    # ignore first event (no image data)
    for ii in range(1, len(ds)):
        image_bg = ds["image_bg"][ii] + bg_off[ii]
        image_corr = np.array(ds["image"][ii], dtype=int) - image_bg
        mask = ds["mask"][ii]
        p10 = np.percentile(image_corr[mask], 10)
        p90 = np.percentile(image_corr[mask], 90)
        assert np.max(np.abs(image_corr)) < 50, "sanity check"
        assert ds["bright_perc_10"][ii] == p10
        assert ds["bright_perc_90"][ii] == p90


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_af_brightness_bc_single():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    ds = dclab.new_dataset(path)
    # sanity checks
    assert "bright_perc_10" not in ds.features_innate
    assert "bright_perc_90" not in ds.features_innate
    # ignore first event (no image data)
    for ii in range(1, len(ds)):
        image_corr = np.array(ds["image"][ii], dtype=int) - ds["image_bg"][ii]
        mask = ds["mask"][ii]
        p10 = np.percentile(image_corr[mask], 10)
        p90 = np.percentile(image_corr[mask], 90)
        assert np.max(np.abs(image_corr)) < 50, "sanity check"

        p10_calc, p90_calc = dclab.features.bright_perc.get_bright_perc(
            ds["mask"][ii],
            ds["image"][ii],
            ds["image_bg"][ii],
        )
        assert (p10_calc, p90_calc) == (p10, p90)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_af_brightness_bc_list():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    ds = dclab.new_dataset(path)
    # sanity checks
    assert "bright_perc_10" not in ds.features_innate
    assert "bright_perc_90" not in ds.features_innate
    # ignore first event (no image data)
    mask = list(ds["mask"][1:6])
    image = list(ds["image"][1:6])
    image_bg = list(ds["image_bg"][1:6])

    p10_calc, p90_calc = dclab.features.bright_perc.get_bright_perc(
        mask, image, image_bg,
    )
    assert isinstance(p10_calc, np.ndarray)
    assert isinstance(p90_calc, np.ndarray)
    assert p10_calc.shape == (4,)
    assert p90_calc.shape == (4,)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_af_brightness_bc_single_with_offset():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    with h5py.File(path, "a") as h5:
        bg_off = np.linspace(0, 1, len(h5["events/deform"]))
        h5["events/bg_off"] = bg_off

    ds = dclab.new_dataset(path)
    # sanity checks
    assert "bright_perc_10" not in ds.features_innate
    assert "bright_perc_90" not in ds.features_innate
    # ignore first event (no image data)
    for ii in range(1, len(ds)):
        image_bg = ds["image_bg"][ii] + bg_off[ii]
        image_corr = np.array(ds["image"][ii], dtype=int) - image_bg
        mask = ds["mask"][ii]
        p10 = np.percentile(image_corr[mask], 10)
        p90 = np.percentile(image_corr[mask], 90)
        assert np.max(np.abs(image_corr)) < 50, "sanity check"
        assert dclab.features.bright_perc.get_bright_perc(
            ds["mask"][ii],
            ds["image"][ii],
            ds["image_bg"][ii],
            ds["bg_off"][ii],
        ) == (p10, p90)



"""Test access to private resources on GitHub"""
import copy
import os
import time

import dclab
from dclab.rtdc_dataset.fmt_dcor import api

from helper_methods import DCOR_AVAILABLE

import pytest

pytest.importorskip("requests")

if not DCOR_AVAILABLE:
    pytest.skip("No connection to DCOR", allow_module_level=True)


# 250209_Blood_2025-02-09_09.46_M003_Reference_dcn_export_28_minimal.rtdc
# a233aaf8-9998-4c44-8070-20fdba7cf3b2
rid = "aa452e24-6088-4bf7-839f-f7989ef38cf9"
access_token = os.environ.get("DCOR_API_TOKEN")


@pytest.mark.skipif(not access_token,
                    reason="DCOR_API_TOKEN environment variable not set")
def test_private_data_access():
    with dclab.new_dataset(rid, api_key=access_token) as ds:
        assert len(ds) == 28


def test_private_data_access_forbidden():
    """This is a control test without an API token"""
    with pytest.raises(api.DCORAccessError, match="Access denied"):
        dclab.new_dataset(rid)


@pytest.mark.skipif(not access_token,
                    reason="DCOR_API_TOKEN environment variable not set")
def test_basin_perishable_repr():
    with dclab.new_dataset(rid, api_key=access_token) as ds:
        bn = ds.basins[0]
        assert "valid" in repr(bn.perishable)


@pytest.mark.skipif(not access_token,
                    reason="DCOR_API_TOKEN environment variable not set")
def test_basin_refresh():
    with dclab.new_dataset(rid, api_key=access_token) as ds:
        ds.cache_basin_dict_time = 0.1
        bn = ds.basins[0]
        ds_bn0 = bn.ds
        assert bn.perishable
        assert not bn.perishable.perished()
        current_kwargs = copy.deepcopy(bn.perishable.expiration_kwargs)
        # Wait longer than `ds.cache_basin_dict_time` so it fetches the basin
        # information anew.
        time.sleep(0.2)
        # Refresh the basin.
        bn.perishable.refresh()
        assert bn.ds is not ds_bn0, "dataset should change"
        # Since the expiration time is computed using
        # "time_local_request" and there is random latency when
        # communicating with the server, the "time_local_expiration" in
        # the `expiration_kwargs` should always be different.
        assert current_kwargs != bn.perishable.expiration_kwargs


@pytest.mark.skipif(not access_token,
                    reason="DCOR_API_TOKEN environment variable not set")
def test_basin_as_dict():
    with dclab.new_dataset(rid, api_key=access_token) as ds:
        ds.cache_basin_dict_time = 0.1
        bn = ds.basins[0]
        bn_dict = bn.as_dict()
        assert bn_dict["perishable"]
        assert isinstance(bn_dict["perishable"], bool)



"""Overview plots for the Young's modulus look-up table

The Young's modulus is corrected for pixelation (pixel size 0.34 µm).
The temperature is set to 23°C.
The 20µm plot is used in the dclab documentation. The others
are used in the DCscope documentation.
"""
from dclab.features import emodulus
import numpy as np
import seaborn as sns
import matplotlib.pylab as plt
from matplotlib import cm, colors
from matplotlib.ticker import MultipleLocator

sns.set_style("whitegrid")

media = ["CellCarrier", "CellCarrier B", "water"]
matrix = {  # channel width: flow rates
    15: [0.016, 0.032, 0.048],
    20: [0.040, 0.080, 0.120],
    30: [0.160, 0.240, 0.320],
    40: [0.320, 0.400, 0.600],
}
ylim = (0, 0.2)
xlim = {  # channel width: xmax
    15: (10, 200),
    20: (20, 350),
    30: (50, 800),
    40: (120, 1400),
}

deform_max = 0.17
grid_size = 1000
emin = 4e-2
emax = 2e1
norm = colors.LogNorm(vmin=emin, vmax=emax)
cmap = "jet"


for lut_name in [
    "LE-2D-FEM-19",
    "HE-2D-FEM-22",
    "HE-3D-FEM-22"
]:

    for channel_width in matrix:
        flow_rates = matrix[channel_width]
        # setup figure
        fig, axes = plt.subplots(nrows=3, ncols=3, sharex=True, sharey=True,
                                 gridspec_kw={"left": 0.12,
                                              "right": .85,
                                              "wspace": .1,
                                              "hspace": .15,
                                              },
                                 figsize=(7.5, 6.3))
        fig.suptitle("{} µm channel".format(channel_width))
        axes[2][1].set_xlabel("Area [µm²]")
        axes[1][0].set_ylabel("Deformation")
        # column label (flow rate)
        for ii in range(3):
            axes[0][ii].set_title("{} µL/s".format(flow_rates[ii]),
                                  fontsize=10, fontweight="bold")
        # row label (medium)
        for jj in range(3):
            pos = axes[jj][0].get_position()
            x = pos.x0 - .11
            y = pos.y0 + pos.height/2
            fig.text(x, y, media[jj], ha='left', va='center', rotation=90,
                     fontsize=10, fontweight="bold")

        # plot the LUTs channel-wise
        lut_base, lut_meta = emodulus.load_lut(lut_name)
        deform = np.linspace(lut_base[:, 1].min(), lut_base[:, 1].max(),
                             grid_size, endpoint=True)
        area_um = np.linspace(lut_base[:, 0].min(), lut_base[:, 0].max(),
                              grid_size, endpoint=True)
        for jj, medium in enumerate(media):  # rows
            for ii, flow_rate in enumerate(matrix[channel_width]):  # columns
                area_um_scaled = emodulus.scale_linear.scale_area_um(
                    area_um=area_um,
                    channel_width_in=lut_meta["channel_width"],
                    channel_width_out=channel_width)
                aa, dd = np.meshgrid(area_um_scaled, deform)
                emod = emodulus.get_emodulus(area_um=aa,
                                             deform=dd,
                                             medium=medium,
                                             channel_width=channel_width,
                                             flow_rate=flow_rate,
                                             px_um=.34,
                                             temperature=23.0,
                                             visc_model="buyukurganci-2022",
                                             lut_data=lut_name)
                ax = axes[jj][ii]
                ax.set_aspect("auto")
                pc = ax.pcolormesh(aa, dd, emod, shading="nearest", cmap=cmap,
                                   norm=norm, zorder=11)
                ax.grid(zorder=10)
                # ticks every .5
                ax.yaxis.set_major_locator(MultipleLocator(0.05))

        axes[2][2].set_ylim(*ylim)
        axes[2][2].set_xlim(*xlim[channel_width])
        axes[2][2].grid(True, which="minor", zorder=10)

        # add colorbar
        cax = plt.axes([0.87, 0.1, 0.02, 0.8])

        mappable = cm.ScalarMappable(norm=norm, cmap=cmap)
        plt.colorbar(pc, orientation="vertical", cax=cax, extend="both",
                     label=f"apparent Young's modulus ({lut_name}) [kPa]")

        plt.savefig("_emodulus_{}um_{}.png".format(channel_width, lut_name),
                    dpi=150)
        plt.close()



import warnings

from .kde.contours import (  # noqa: F401
    find_contours_level, _find_quantile_level, get_quantile_levels
)


warnings.warn("`dclab.kde_contours` is deprecated; please use "
              "the `dclab.kde.contours` instead",
              DeprecationWarning)



import pytest
import numpy as np
import dclab


def test_kde_contours_deprecated_warning():
    with pytest.deprecated_call():
        np.random.seed(47)
        x0 = np.random.normal(loc=100, scale=10, size=1000)
        y0 = np.random.normal(loc=.1, scale=.01, size=1000)

        ds = dclab.new_dataset({"area_um": x0, "deform": y0})
        ds.config["filtering"]["enable filters"] = False

        x, y, kde = ds.get_kde_contour(xax="area_um",
                                       yax="deform",
                                       xacc=.10,
                                       yacc=.01,
                                       kde_type="histogram")
        level = dclab.kde_contours.get_quantile_levels(
            density=kde,
            x=x,
            y=y,
            xp=ds["area_um"],
            yp=ds["deform"],
            q=.89,
            normalize=True
        )

        level2, err = dclab.kde_contours._find_quantile_level(
            density=kde,
            x=x,
            y=y,
            xp=ds["area_um"],
            yp=ds["deform"],
            quantile=.89,
            acc=0,
            ret_err=True
        )
        # since _find_quantile level does not do linear interpolation
        # in the density, the computed values can differ from the values
        # obtained using get_quantile_levels - even with err==0.
        assert err == 0
        # This is the resulting level difference.
        assert np.abs(level - level2) < 0.00116


def test_kde_methods_deprecated_warning():
    with pytest.deprecated_call():
        a = np.arange(100)
        b = dclab.kde_methods.bin_width_doane(a)
        assert np.allclose(b, 12.951578044133464)



import numpy as np
import dclab.kde.methods


def test_bin_width_doane():
    a = np.arange(100)
    b = dclab.kde.methods.bin_width_doane(a)
    assert np.allclose(b, 12.951578044133464)


def test_bin_width_percentile():
    a = np.arange(100)
    b = dclab.kde.methods.bin_width_percentile(a)
    assert np.allclose(b, 3.4434782608695653)



"""Volume computation based on contour revolution"""
from __future__ import annotations

import numpy as np
import numpy.typing as npt


def get_volume(
        cont: npt.NDArray | list[npt.NDArray],
        pos_x: float | npt.NDArray,
        pos_y: float | npt.NDArray,
        pix: float,
        fix_orientation: bool = False) -> float | npt.NDArray:
    """Calculate the volume of a polygon revolved around an axis

    The volume estimation assumes rotational symmetry.

    Parameters
    ----------
    cont: ndarray or list of ndarrays of shape (N,2)
        A 2D array that holds the contour of an event [px]
        e.g. obtained using `mm.contour` where  `mm` is an instance
        of `RTDCBase`. The first and second columns of `cont`
        correspond to the x- and y-coordinates of the contour.
    pos_x: float or ndarray of length N
        The x coordinate(s) of the centroid of the event(s) [µm]
        e.g. obtained using `mm.pos_x`
    pos_y: float or ndarray of length N
        The y coordinate(s) of the centroid of the event(s) [µm]
        e.g. obtained using `mm.pos_y`
    pix
        The detector pixel size in µm.
        e.g. obtained using: `mm.config["imaging"]["pixel size"]`
    fix_orientation
        If set to True, make sure that the orientation of the
        contour is counter-clockwise in the r-z plane
        (see :func:`vol_revolve`). This is False by default, because
        (1) Shape-In always stores the contours in the correct
        orientation and (2) there may be events with high porosity
        where "fixing" the orientation makes things worse and a
        negative volume is returned.

    Returns
    -------
    volume
        volume in um^3

    Notes
    -----
    The computation of the volume is based on a full rotation of the
    upper and the lower halves of the contour from which the
    average is then used.

    The volume is computed radially from the center position
    given by (`pos_x`, `pos_y`). For sufficiently smooth contours,
    such as densely sampled ellipses, the center position does not
    play an important role. For contours that are given on a coarse
    grid, as is the case for RT-DC, the center position must be
    given.

    References
    ----------
    - https://de.wikipedia.org/wiki/Kegelstumpf#Formeln
    - Yields identical results to the Matlab script by Geoff Olynyk
      <https://de.mathworks.com/matlabcentral/fileexchange/36525-volrevolve>`_
    """
    if np.isscalar(pos_x):
        cont = [cont]
        ret_list = False
    else:
        ret_list = True

    # Convert input to 1D arrays
    pos_x = np.atleast_1d(pos_x)
    pos_y = np.atleast_1d(pos_y)

    if pos_x.size != pos_y.size:
        raise ValueError("Size of `pos_x` and `pos_y` must match!")

    if pos_x.size > 1 and len(cont) <= 1:
        raise ValueError("Number of given contours too small!")

    # results are stored in a separate array initialized with nans
    v_avg = np.zeros_like(pos_x, dtype=np.float64) * np.nan

    # v_avg has the shape of `pos_x`. We are iterating over the smallest
    # length for `cont` and `pos_x`.
    for ii in range(min(len(cont), pos_x.shape[0])):
        # If the contour has less than 4 pixels, the computation will fail.
        # In that case, the value np.nan is already assigned.
        cc = cont[ii]
        if cc.shape[0] >= 4:
            # Center contour coordinates with given centroid
            contour_x = cc[:, 0] - pos_x[ii] / pix
            contour_y = cc[:, 1] - pos_y[ii] / pix
            # Switch to r and z to follow notation of vol_revolve
            # (In RT-DC the axis of rotation is x, but for vol_revolve
            # we need the axis vertically)
            contour_r = contour_y
            contour_z = contour_x
            if fix_orientation:
                # Make sure the contour is counter-clockwise
                contour_r, contour_z = counter_clockwise(contour_r, contour_z)

            # Compute right volume
            # Which points are at negative r-values (r<0)?
            inx_neg = np.where(contour_r < 0)
            # These points will be shifted up to r=0 directly on the z-axis
            contour_right = np.copy(contour_r)
            contour_right[inx_neg] = 0
            vol_right = vol_revolve(contour_right, contour_x, pix)

            # Compute left volume
            # Which points are at positive r-values? (r>0)?
            idx_pos = np.where(contour_r > 0)
            # These points will be shifted down to y=0 to build an x-axis
            contour_left = np.copy(contour_r)
            contour_left[idx_pos] = 0
            # Now we still have negative r values, but vol_revolve needs
            # positive values, so we flip the sign...
            contour_left[:] *= -1
            # ... but in doing so, we have switched to clockwise rotation
            # and we need to pass the array in reverse order
            vol_left = vol_revolve(contour_left[::-1], contour_x[::-1], pix)

            # Compute the average
            v_avg[ii] = (vol_right + vol_left) / 2

    if not ret_list:
        # Do not return a list if the input contour was not in a list
        v_avg = v_avg[0]

    return v_avg


def counter_clockwise(cx: npt.NDArray, cy: npt.NDArray) -> tuple[float, float]:
    """Put contour coordinates into counter-clockwise order

    Parameters
    ----------
    cx, cy: 1d ndarrays
        The x- and y-coordinates of the contour

    Returns
    -------
    cx_cc, cy_cc
        The x- and y-coordinates of the contour in
        counter-clockwise orientation.

    Notes
    -----
    The contour must be centered around (0, 0).
    """
    # test orientation
    angles = np.unwrap(np.arctan2(cy, cx))
    grad = np.diff(angles)
    if np.average(grad) < 0:
        return cx[::-1], cy[::-1]
    else:
        return cx, cy


def vol_revolve(r: npt.NDArray,
                z: npt.NDArray,
                point_scale: float = 1.) -> float | npt.NDArray:
    r"""Calculate the volume of a polygon revolved around the Z-axis

    This implementation yields the same results as the volRevolve
    Matlab function by Geoff Olynyk (from 2012-05-03)
    https://de.mathworks.com/matlabcentral/fileexchange/36525-volrevolve.

    The difference here is that the volume is computed using (a much
    more approachable) implementation using the volume of a truncated
    cone (https://de.wikipedia.org/wiki/Kegelstumpf).

    .. math::

      V = \frac{h \cdot \pi}{3} \cdot (R^2 + R \cdot r + r^2)

    Where :math:`h` is the height of the cone and :math:`r` and
    `R` are the smaller and larger radii of the truncated cone.

    Each line segment of the contour resembles one truncated cone. If
    the z-step is positive (counter-clockwise contour), then the
    truncated cone volume is added to the total volume. If the z-step
    is negative (e.g. inclusion), then the truncated cone volume is
    removed from the total volume.

    .. versionchanged:: 0.37.0

      The volume in previous versions was overestimated by on average
      2µm³.

    Parameters
    ----------
    r: 1d ndarray
        radial coordinates (perpendicular to the z axis)
    z: 1d ndarray
        coordinate along the axis of rotation
    point_scale
        point size in your preferred units; The volume is multiplied
        by a factor of `point_scale**3`.

    Notes
    -----
    The coordinates must be given in counter-clockwise order,
    otherwise the volume will be negative.
    """
    r = np.array(r).flatten()
    z = np.array(z).flatten()

    # sanity checks
    assert len(r) == len(z)
    assert len(r) >= 3
    assert len(r.shape) == len(z.shape) == 1
    assert np.all(r >= 0)

    # make sure we have a closed contour
    if (r[-1] != r[0]) or (z[-1] != z[0]):
        # We have an open contour - close it.
        r = np.resize(r, len(r) + 1)
        z = np.resize(z, len(z) + 1)

    rp = r[:-1]

    # array of radii differences: R - r
    dr = np.diff(r)
    # array of height differences: h
    dz = np.diff(z)

    # If we expand the function in the doc string with
    # dr = R - r and dz = h, then we get three terms for the volume
    # (as opposed to four terms in Olynyk's script). Those three terms
    # all resemble area slices multiplied by the z-distance dz.
    a1 = 3 * rp ** 2
    a2 = 3 * rp * dr
    a3 = dr ** 2

    # Note that the formula for computing the volume is symmetric
    # with respect to r and R. This means that it does not matter
    # which sign dr has (R and r are always positive). Since this
    # algorithm assumes that the contour is ordered counter-clockwise,
    # positive dz means adding to the contour while negative dz means
    # subtracting from the contour (see test functions for more details).
    # Conveniently so, dz only appears one time in this formula, so
    # we can take the sign of dz as it is (Otherwise, we would have
    # to take the absolute value of every truncated cone volume and
    # multiply it by np.sign(dz)).
    v = np.pi / 3 * dz * np.abs(a1 + a2 + a3)
    vol = np.sum(v) * point_scale ** 3

    return vol



"""Test HTTP format"""
import time

import numpy as np
import pytest


from dclab.rtdc_dataset.fmt_http import RTDC_HTTP

from helper_methods import DCOR_AVAILABLE


pytest.importorskip("requests")

if not DCOR_AVAILABLE:
    pytest.skip("No connection to DCOR", allow_module_level=True)

# 250209_Blood_2025-02-09_09.46_M003_Reference_dcn_export_28.rtdc
s3_url = ("https://objectstore.hpccloud.mpcdf.mpg.de/"
          "circle-442e6d53-c48b-46eb-873c-4a0f98f3827d/"
          "resource/57e/cde/5d-f896-4599-ba35-d1be7defc6fe")


def test_cache_features():
    with RTDC_HTTP(s3_url) as ds:
        t0 = time.perf_counter()
        _ = ds["deform"][:]
        _ = ds["image"][10]
        t1 = time.perf_counter()
        for ii in range(50):
            _ = ds["deform"][:]
            _ = ds["image"][10]
        t2 = time.perf_counter()
        assert t2 - t1 < t1 - t0


def test_identifier():
    with RTDC_HTTP(s3_url) as ds:
        # This is the HTTP ETag (https://en.wikipedia.org/wiki/HTTP_ETag)
        # given to this resource by the object store. If the file is
        # re-uploaded, the ETag may change and this test will fail.
        assert ds.identifier == "6dd392feb1aeda7cfb73b4ec76c1fe7c"


@pytest.mark.parametrize("netloc", [
    "objectstore.hpccloud.mpcdf.mpg.de",
    "objectstore.hpccloud.mpcdf.mpg.de:443"
])
def test_netloc_vs_hostname(netloc):
    with RTDC_HTTP(s3_url) as ds:
        assert len(ds) == 28
        assert np.allclose(ds["deform"][0], 0.05335504858810891,
                           rtol=0, atol=1e-7)


def test_open_public_s3_dataset():
    with RTDC_HTTP(s3_url) as ds:
        assert ds.config["experiment"]["sample"] == "Reference"
        assert len(ds) == 28
        assert np.allclose(ds["deform"][0], 0.05335504858810891,
                           atol=0,
                           rtol=1e-5)



"""Core tools for the analysis of deformability cytometry datasets

Copyright (C) 2015 Paul Müller

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
# flake8: noqa: F401
from . import definitions as dfn
from . import features
from . import isoelastics
from . import lme4
from .polygon_filter import PolygonFilter
from . import rtdc_dataset
from .rtdc_dataset import new_dataset, IntegrityChecker, RTDCWriter
from .rtdc_dataset.feat_temp import (
    register_temporary_feature, set_temporary_feature)
from .rtdc_dataset.feat_anc_ml import (
    MachineLearningFeature, load_modc, load_ml_feature, save_modc)
from .rtdc_dataset.feat_anc_plugin.plugin_feature import (
    PlugInFeature, load_plugin_feature)
from . import statistics
from . import util

from ._version import __version__, __version_tuple__


# Lazy-load deprecated kde modules 
kde_contours = util.LazyLoader("dclab.kde_contours")
kde_methods = util.LazyLoader("dclab.kde_methods")



import pytest

import dclab
from dclab.definitions import get_feature_label


@pytest.mark.parametrize("name,label",
                         [
                             ["ml_score_rbc", "ML score RBC"],
                             ["deform", "Deformation"],
                             ["area_um", "Area [µm²]"],
                         ])
def test_get_feature_label_basic(name, label):
    assert get_feature_label(name) == label


def test_get_feature_label_temporary():
    name = "unknown"
    label = "User-defined feature unknown"
    dclab.register_temporary_feature(name)
    assert get_feature_label(name) == label


def test_get_feature_label_invalid():
    name = "unknown2"
    with pytest.raises(ValueError, match="unknown2"):
        get_feature_label(name)



"""Export RT-DC measurement data"""
from __future__ import annotations

import codecs
import json
import pathlib
import time
from typing import Dict, List
import uuid
import warnings

import h5py
import hdf5plugin

try:
    import av
except ModuleNotFoundError:
    PYAV_AVAILABLE = False
else:
    PYAV_AVAILABLE = True

try:
    import fcswrite
except ModuleNotFoundError:
    FCSWRITE_AVAILABLE = False
else:
    FCSWRITE_AVAILABLE = True

import numpy as np

from .. import definitions as dfn
from .._version import version, version_tuple

from .feat_basin import get_basin_classes
from .writer import RTDCWriter


class LimitingExportSizeWarning(UserWarning):
    pass


class ContourNotExportedWarning(UserWarning):
    pass


class Export(object):
    def __init__(self, rtdc_ds):
        """Export functionalities for RT-DC datasets"""
        self.rtdc_ds = rtdc_ds

    def avi(self,
            path: str | pathlib.Path,
            filtered: bool = True,
            override: bool = False,
            pixel_format: str = "yuv420p",
            codec: str = "rawvideo",
            codec_options: dict[str, str] = None,
            progress_callback: callable = None,
            ):
        """Exports filtered event images to a video file

        Parameters
        ----------
        path: str
            Path to a video file. The container is (.avi, .mkv, ...) is
            deduced from the file suffix.
        filtered: bool
            If set to `True`, only the filtered data
            (index in ds.filter.all) are used.
        override: bool
            If set to `True`, an existing file ``path`` will be overridden.
            If set to `False`, raises `OSError` if ``path`` exists.
        pixel_format: str
            Which pixel format to give to ffmpeg.
        codec: str
            Codec name; e.g. "rawvideo" or "libx264"
        codec_options:
            Additional arguments to give to the codec using ffmpeg,
            e.g. `{'preset': 'slow', 'crf': '0'}` for "libx264" codec.
        progress_callback: callable
            Function that takes at least two arguments: float between 0 and
            1 for monitoring progress and a string describing what is being
            done.

        Notes
        -----
        Raises OSError if current dataset does not contain image data
        """
        if not PYAV_AVAILABLE:
            raise ModuleNotFoundError(
                "Package `av` required for avi export!")
        path = pathlib.Path(path)
        if len(path.suffix) != 4:
            path = path.with_suffix(".avi")
        ds = self.rtdc_ds
        # Check if file already exist
        if not override and path.exists():
            raise OSError("File already exists: {}\n".format(
                str(path).encode("ascii", "ignore")) +
                "Please use the `override=True` option.")
        # Start exporting
        if "image" in ds:
            # Open video for writing
            with av.open(path, mode="w") as container:
                stream = container.add_stream(codec_name=codec,
                                              rate=25)
                stream.pix_fmt = pixel_format
                stream.height = ds["image"].shape[1]
                stream.width = ds["image"].shape[2]
                if codec_options:
                    stream.codec_context.options = codec_options

                # write the filtered frames to the video file
                for evid in np.arange(len(ds)):

                    if progress_callback is not None and evid % 10_000 == 0:
                        progress_callback(evid / len(ds), "exporting video")

                    # skip frames that were filtered out
                    if filtered and not ds.filter.all[evid]:
                        continue
                    image = ds["image"][evid]
                    # Convert to RGB
                    image = image.reshape(image.shape[0], image.shape[1], 1)
                    image = np.repeat(image, 3, axis=2)

                    av_frame = av.VideoFrame.from_ndarray(image,
                                                          format="rgb24")

                    for packet in stream.encode(av_frame):
                        container.mux(packet)

            if progress_callback is not None:
                progress_callback(1.0, "video export complete")

        else:
            msg = "No image data to export: dataset {} !".format(ds.title)
            raise OSError(msg)

    def fcs(self,
            path: pathlib.Path | str,
            features: list[str],
            meta_data: dict = None,
            filtered: bool = True,
            override: bool = False,
            progress_callback: callable = None,
            ):
        """Export the data of an RT-DC dataset to an .fcs file

        Parameters
        ----------
        path: str
            Path to an .fcs file. The ending .fcs is added automatically.
        features: list of str
            The features in the resulting .fcs file. These are strings
            that are defined by `dclab.definitions.scalar_feature_exists`,
            e.g. "area_cvx", "deform", "frame", "fl1_max", "aspect".
        meta_data: dict
            User-defined, optional key-value pairs that are stored
            in the primary TEXT segment of the FCS file; the version
            of dclab is stored there by default
        filtered: bool
            If set to `True`, only the filtered data
            (index in ds.filter.all) are used.
        override: bool
            If set to `True`, an existing file ``path`` will be overridden.
            If set to `False`, raises `OSError` if ``path`` exists.
        progress_callback: callable
            Function that takes at least two arguments: float between 0 and
            1 for monitoring progress and a string describing what is being
            done.

        Notes
        -----
        Due to incompatibility with the .fcs file format, all events with
        NaN-valued features are not exported.
        """
        if meta_data is None:
            meta_data = {}
        if not FCSWRITE_AVAILABLE:
            raise ModuleNotFoundError(
                "Package `fcswrite` required for fcs export!")

        ds = self.rtdc_ds

        path = pathlib.Path(path)
        # Make sure that path ends with .fcs
        if path.suffix != ".fcs":
            path = path.with_name(path.name + ".fcs")
        # Check if file already exist
        if not override and path.exists():
            raise OSError("File already exists: {}\n".format(
                str(path).encode("ascii", "ignore")) +
                "Please use the `override=True` option.")
        # Check that features are valid
        features = sorted(set(features))
        for c in features:
            if c not in ds.features_scalar:
                msg = "Invalid feature name: {}".format(c)
                raise ValueError(msg)

        # Collect the header
        chn_names = [dfn.get_feature_label(c, rtdc_ds=ds) for c in features]

        if progress_callback is not None:
            progress_callback(0.0, "collecting data")

        # Collect the data
        if filtered:
            data = [ds[c][ds.filter.all] for c in features]
        else:
            data = [ds[c] for c in features]

        if progress_callback is not None:
            progress_callback(0.5, "exporting data")

        data = np.array(data).transpose()
        meta_data["dclab version"] = version
        fcswrite.write_fcs(filename=str(path),
                           chn_names=chn_names,
                           data=data,
                           text_kw_pr=meta_data,
                           )

        if progress_callback is not None:
            progress_callback(1.0, "export complete")

    def hdf5(self,
             path: str | pathlib.Path,
             features: List[str] = None,
             filtered: bool = True,
             logs: bool = False,
             tables: bool = False,
             basins: bool = False,
             allow_contour: bool = False,
             meta_prefix: str = "src_",
             override: bool = False,
             compression_kwargs: Dict = None,
             compression: str = "deprecated",
             skip_checks: bool = False,
             progress_callback: callable = None,
             ):
        """Export the data of the current instance to an HDF5 file

        Parameters
        ----------
        path: str
            Path to an .rtdc file. The ending .rtdc is added
            automatically.
        features: list of str
            The features in the resulting .rtdc file. These are strings
            that are defined by `dclab.definitions.feature_exists`, e.g.
            "area_cvx", "deform", "frame", "fl1_max", "image".
            Defaults to `self.rtdc_ds.features_innate`.
        filtered: bool
            If set to `True`, only the filtered data
            (index in ds.filter.all) are used.
        logs: bool
            Whether to store the logs of the original file prefixed with
            `source_` to the output file.
        tables: bool
            Whether to store the tables of the original file prefixed with
            `source_` to the output file.
        basins: bool
            Whether to export basins. If filtering is disabled, basins
            are copied directly to the output file. If filtering is enabled,
            then mapped basins are exported.
        allow_contour: bool
            Whether to allow exporting the "contour" feature. Writing this
            feature to an HDF5 file is extremely inefficient, because it
            cannot be represented by an ND array and thus must be stored
            in a group, each contour stored in a separate dataset. The
            contour can easily be computed via the mask, so actually storing
            the contour should be avoided. If "contour" is in `features`,
            it will only be written to the output file if `allow_contour=True`.
        meta_prefix: str
            Prefix for log and table names in the exported file
        override: bool
            If set to `True`, an existing file ``path`` will be overridden.
            If set to `False`, raises `OSError` if ``path`` exists.
        compression_kwargs: dict
            Dictionary with the keys "compression" and "compression_opts"
            which are passed to :func:`h5py.H5File.create_dataset`. The
            default is Zstandard compression with the compression
            level 5 `hdf5plugin.Zstd(clevel=5)`.
        compression: str or None
            Compression method used for data storage;
            one of [None, "lzf", "gzip", "szip"].

            .. deprecated:: 0.43.0
                Use `compression_kwargs` instead.
        skip_checks: bool
            Disable checking whether all features have the same length.
        progress_callback: callable
            Function that takes at least two arguments: float between 0 and
            1 for monitoring progress and a string describing what is being
            done.

        .. versionchanged:: 0.58.0

           The ``basins`` keyword argument was added, and it is now possible
           to pass an empty list to ``features``. This combination results
           in a very small file consisting of metadata and a mapped basin
           referring to the original dataset.
        """
        if compression != "deprecated":
            warnings.warn("The `compression` kwarg is deprecated in favor of "
                          "`compression_kwargs`!",
                          DeprecationWarning)
            if compression_kwargs is not None:
                raise ValueError("You may not specify `compression` and "
                                 "`compression_kwargs` at the same time!")
            # be backwards-compatible
            compression_kwargs = {"compression": compression}
        if compression_kwargs is None:
            compression_kwargs = hdf5plugin.Zstd(clevel=5)
        path = pathlib.Path(path)
        # Make sure that path ends with .rtdc
        if path.suffix not in [".rtdc", ".rtdc~"]:
            path = path.parent / (path.name + ".rtdc")
        # Check if file already exists
        if not override and path.exists():
            raise OSError("File already exists: {}\n".format(path)
                          + "Please use the `override=True` option.")
        elif path.exists():
            path.unlink()

        # make sure the parent directory exists
        path.parent.mkdir(parents=True, exist_ok=True)

        # for convenience
        ds = self.rtdc_ds

        # remove contour information from user-specified features
        if "contour" in (features or []) and not allow_contour:
            features = list(features)
            features.remove("contour")
            warnings.warn(
                "Feature 'contour' not exported to output file, because "
                "`allow_contour` is `False`. If you really need the "
                "'contour' feature in the output file (unlikely, unless you "
                "are venturing outside the DC Cosmos), you must set "
                "`allow_contour=True`. Otherwise, you can safely ignore "
                "this warning or silence it by not providing 'contour' in "
                "`features`.",
                ContourNotExportedWarning)

        if features is None:
            features = ds.features_innate
            # silently remove contour information
            if "contour" in features and not allow_contour:
                features.remove("contour")

        # decide which metadata to export
        meta = {}
        # only cfg metadata (no analysis metadata)
        for sec in dfn.CFG_METADATA:
            if sec in ds.config:
                meta[sec] = ds.config[sec].copy()
        # add user-defined metadata
        if "user" in ds.config:
            meta["user"] = ds.config["user"].copy()
        if filtered:
            # Define a new measurement identifier, so that we are not running
            # into any problems with basins being defined for filtered data.
            ds_run_id = ds.get_measurement_identifier()
            random_ap = f"dclab-{str(uuid.uuid4())[:7]}"
            meta["experiment"]["run identifier"] = f"{ds_run_id}_{random_ap}"

        if filtered:
            filter_arr = ds.filter.all
        else:
            filter_arr = None

        features = sorted(set(features))
        if not skip_checks and features:
            # check that all features have same length and use the smallest
            # common length
            lengths = []
            for feat in features:
                if feat == "trace":
                    for tr in list(ds["trace"].keys()):
                        lengths.append(len(ds["trace"][tr]))
                else:
                    lengths.append(len(ds[feat]))
            l_min = np.min(lengths)
            l_max = np.max(lengths)
            if l_min != l_max:
                if filter_arr is None:
                    # we are forced to do filtering
                    filter_arr = np.ones(len(ds), dtype=bool)
                else:
                    # have to create a copy, because rtdc_ds.filter.all is ro!
                    filter_arr = np.copy(filter_arr)
                filter_arr[l_min:] = False
                warnings.warn(
                    "Not all features have the same length! Limiting output "
                    + f"event count to {l_min} (max {l_max}) in '{l_min}'.",
                    LimitingExportSizeWarning)

        # Perform actual export
        with RTDCWriter(path,
                        mode="append",
                        compression_kwargs=compression_kwargs) as hw:
            if progress_callback is not None:
                progress_callback(0.0, "writing metadata")
            # write meta data
            hw.store_metadata(meta)

            # write export log
            hw.store_log(time.strftime("dclab-export_%Y-%m-%d_%H.%M.%S"),
                         json.dumps(
                             {"dclab version": version_tuple,
                              "kwargs": {
                                  "features": features,
                                  "filtered": filtered,
                                  "logs": logs,
                                  "tables": tables,
                                  "basins": basins,
                                  "meta_prefix": meta_prefix,
                                  "skip_checks": skip_checks
                                  }
                              },
                             indent=2,
                             sort_keys=True,
                         ).split("\n"))

            if logs:
                # write logs
                for log in ds.logs:
                    hw.store_log(f"{meta_prefix}{log}",
                                 ds.logs[log])

            if tables:
                # write tables
                for tab in ds.tables:
                    hw.store_table(f"{meta_prefix}{tab}",
                                   ds.tables[tab])

            # write each feature individually
            for ii, feat in enumerate(features):
                if progress_callback is not None:
                    progress_callback(ii / len(features), f"exporting {feat}")

                if (filter_arr is None or
                        # This does not work for the .tdms file format
                        # (and probably also not for DCOR).
                        (np.all(filter_arr) and ds.format == "hdf5")):
                    # We do not have to filter and can be fast
                    if dfn.scalar_feature_exists(feat):
                        shape = (1,)
                    elif feat in ["image", "image_bg", "mask", "trace"]:
                        # known shape
                        shape = None
                    else:
                        shape = np.array(ds[feat][0]).shape
                    hw.store_feature(feat=feat,
                                     data=ds[feat],
                                     shape=shape)
                else:
                    # We have to filter and will be slower
                    store_filtered_feature(rtdc_writer=hw,
                                           feat=feat,
                                           data=ds[feat],
                                           filtarr=filter_arr)

            if basins:
                if progress_callback:
                    progress_callback(1 - 1 / (len(features) or 1),
                                      "writing basins")
                # We have to store basins. There are three options:
                # - filtering disabled: just copy basins
                # - filtering enabled
                #   - basins with "same" mapping: create new mapping
                #   - mapped basins: correct nested mapping
                # In addition to the basins that we copy from the
                # original dataset, we also create a new basin that
                # refers to the original dataset itself.
                basin_list = [bn.as_dict() for bn in ds.basins]
                # In addition to the upstream basins, also store a reference
                # to the original file from which the export was done.
                # Get the identifier of the current dataset for the new basins.
                basin_id = ds.get_measurement_identifier()
                if ds.format in get_basin_classes():
                    # The dataset has a format that matches a basin format
                    # directly.
                    basin_is_local = ds.format == "hdf5"
                    basin_locs = [ds.path]
                    if basin_is_local:
                        # So the user can put them into the same directory.
                        basin_locs.append(ds.path.name)
                    basin_list.append({
                        "basin_name": "Exported data",
                        "basin_type": "file" if basin_is_local else "remote",
                        "basin_format": ds.format,
                        "basin_locs": basin_locs,
                        "basin_descr": f"Exported with dclab {version}",
                        "basin_id": basin_id,
                    })
                elif (ds.format == "hierarchy"
                      and ds.get_root_parent().format in get_basin_classes()):
                    # The dataset is a hierarchy child, and it is derived
                    # from a dataset that has a matching basin format.
                    #
                    # For the sake of clarity I wrote this as a separate case,
                    # even if that means duplicating code from the previous
                    # case.
                    ds_root = ds.get_root_parent()
                    basin_is_local = ds_root.format == "hdf5"
                    basin_locs = [ds_root.path]
                    if basin_is_local:
                        # So the user can put them into the same directory.
                        basin_locs.append(ds_root.path.name)
                    basin_list.append({
                        "basin_name": "Exported data (hierarchy)",
                        "basin_type": "file" if basin_is_local else "remote",
                        "basin_format": ds_root.format,
                        "basin_locs": basin_locs,
                        "basin_descr": f"Exported with dclab {version} from a "
                                       f"hierarchy dataset",
                        # Here we do not yet treat the conversion from the
                        # root dataset indices to the child indices,
                        # because we will fill in the missing values below
                        # in the basin mapping correction step.
                        "basin_map": None,
                        "basin_id": basin_id,
                    })

                for bn_dict in basin_list:
                    if bn_dict.get("basin_format") not in get_basin_classes():
                        # Whichever software stored this basin in the
                        # original file, we do not support it or don't want
                        # to break it.
                        continue
                    elif bn_dict.get("basin_type") == "internal":
                        # Internal basins are only valid for files they were
                        # defined in. Since we are exporting, it does not
                        # make sense to store these basins in the output file.
                        continue
                    elif bn_dict.get("perishable"):
                        # Perishable basins require secret keys or complicated
                        # logic to execute in order to refresh them. We do not
                        # store them in the output file.
                        continue

                    # Basin mapping correction: If we are filtering, or
                    # if we are exporting from a hierarchy dataset, we have
                    # to correct or add basin mapping arrays.
                    basinmap_orig = bn_dict.get("basin_map")
                    if ds.format == "hierarchy":
                        # Hierarchy dataset
                        # Compute mapping from hierarchy root.
                        from .fmt_hierarchy import map_indices_child2root
                        map_root = map_indices_child2root(
                            child=ds,
                            child_indices=np.arange(len(ds))
                        )

                        if not filtered and basinmap_orig is None:
                            # We only have to consider the hierarchy.
                            bn_dict["basin_map"] = map_root
                        elif filtered and basinmap_orig is None:
                            # Filtering must be taken into account.
                            bn_dict["basin_map"] = map_root[filter_arr]
                        else:
                            # The source file has mapping defined which we
                            # have to take into account.
                            map_child = basinmap_orig[map_root]
                            if filtered:
                                # Subsetting additional filters
                                bn_dict["basin_map"] = map_child[filter_arr]
                            else:
                                bn_dict["basin_map"] = map_child
                    else:
                        if not filtered:
                            # filtering disabled: just copy basins
                            pass
                        elif filtered and basinmap_orig is None:
                            # basins with mapping "same": create new mapping
                            bn_dict["basin_map"] = np.where(filter_arr)[0]
                        else:
                            # filter the source mapping
                            bn_dict["basin_map"] = basinmap_orig[filter_arr]

                    # Do not verify basins, it takes too long.
                    hw.store_basin(**bn_dict, verify=False)
        if progress_callback is not None:
            progress_callback(1.0, "export complete")

    def tsv(self,
            path: pathlib.Path | str,
            features: list[str],
            meta_data: dict = None,
            filtered: bool = True,
            override: bool = False,
            progress_callback: callable = None,
            ):
        """Export the data of the current instance to a .tsv file

        Parameters
        ----------
        path: str
            Path to a .tsv file. The ending .tsv is added automatically.
        features: list of str
            The features in the resulting .tsv file. These are strings
            that are defined by `dclab.definitions.scalar_feature_exists`,
            e.g. "area_cvx", "deform", "frame", "fl1_max", "aspect".
        meta_data: dict
            User-defined, optional key-value pairs that are stored
            at the beginning of the tsv file - one key-value pair is
            stored per line which starts with a hash. The version of
            dclab is stored there by default.
        filtered: bool
            If set to `True`, only the filtered data
            (index in ds.filter.all) are used.
        override: bool
            If set to `True`, an existing file ``path`` will be overridden.
            If set to `False`, raises `OSError` if ``path`` exists.
        progress_callback: callable
            Function that takes at least two arguments: float between 0 and
            1 for monitoring progress and a string describing what is being
            done.
        """
        if meta_data is None:
            meta_data = {}
        features = [c.lower() for c in features]
        features = sorted(set(features))
        path = pathlib.Path(path)
        ds = self.rtdc_ds
        # Make sure that path ends with .tsv
        if path.suffix != ".tsv":
            path = path.with_name(path.name + ".tsv")
        # Check if file already exist
        if not override and path.exists():
            raise OSError("File already exists: {}\n".format(
                str(path).encode("ascii", "ignore")) +
                "Please use the `override=True` option.")
        # Check that features exist
        for c in features:
            if c not in ds.features_scalar:
                raise ValueError("Invalid feature name {}".format(c))
        meta_data["dclab version"] = version

        if progress_callback is not None:
            progress_callback(0.0, "writing metadata")

        # Write BOM header
        with path.open("wb") as fd:
            fd.write(codecs.BOM_UTF8)
        # Open file
        with path.open("a", encoding="utf-8") as fd:
            # write meta data
            for key in sorted(meta_data.keys()):
                fd.write(f"# {key}: {meta_data[key]}\n")
            fd.write("#\n")
            fd.write("# Original dataset configuration:\n")
            cfg = self.rtdc_ds.config.as_dict()
            for sec in sorted(cfg.keys()):
                for key in sorted(cfg[sec].keys()):
                    fd.write(f"# dc:{sec}:{key} = {cfg[sec][key]}\n")
            fd.write("#\n")
            # write header
            header1 = "\t".join([c for c in features])
            fd.write("# "+header1+"\n")
            labels = [dfn.get_feature_label(c, rtdc_ds=ds) for c in features]
            header2 = "\t".join(labels)
            fd.write("# "+header2+"\n")

        with path.open("ab") as fd:
            if progress_callback is not None:
                progress_callback(0.1, "collecting data")

            # collect data
            if filtered:
                data = [ds[c][ds.filter.all] for c in features]
            else:
                data = [ds[c] for c in features]

            if progress_callback is not None:
                progress_callback(0.5, "writing data")

            np.savetxt(fd,
                       np.array(data).transpose(),
                       fmt=str("%.10e"),
                       delimiter="\t")

        if progress_callback is not None:
            progress_callback(1.0, "export complete")


def yield_filtered_array_stacks(data, indices):
    """Generator returning chunks with the filtered feature data

    Parameters
    ----------
    data: np.ndarray or h5py.Dataset
        The full, unfiltered input feature data. Must implement
        the `shape` and `dtype` properties. If it implements the
        `__array__` method, fast slicing is used.
    indices: np.ndarray or list
        The indices (integer values) for `data` (first axis), indicating
        which elements should be returned by this generator.

    Notes
    -----
    This method works with any feature dimension (e.g. it
    works for image (2D) data and for trace data (1D)). It
    is just important that `data` is indexable using integers
    and that the events in `data` all have the same shape.
    The dtype of the returned chunks is determined by the first
    item in `data`.

    This method works with sliceable (e.g. np.ndarray) and
    non-sliceable (e.g. tdms-format-based images) input data. If the
    input data is sliceable (which is determined by the availability
    of the `__array__` method, then fast numpy sclicing is used. If the
    input data does not support slicing (`__array__` not defined), then
    a slow iteration over `indices` is done.

    In the slow iteration case, the returned array data are overridden
    in-place. If you need to retain a copy of the `yield`ed chunks,
    apply `np.array(.., copy=True)` to the returned chunks.
    """
    chunk_shape = RTDCWriter.get_best_nd_chunks(item_shape=data.shape[1:],
                                                item_dtype=data.dtype)
    chunk_size = chunk_shape[0]

    if hasattr(data, "__array__"):
        # We have an array-like object and can do slicing with the indexing
        # array. This speeds up chunk creation for e.g. the HDF5 file format
        # where all data are present in an array-like fashion.
        indices = np.array(indices)
        stop = 0
        for kk in range(len(indices) // chunk_size):
            start = chunk_size * kk
            stop = chunk_size * (kk + 1)
            yield data[indices[start:stop]]
        if stop < len(indices):
            yield data[indices[stop:]]
    else:
        # assemble filtered image stacks
        chunk = np.zeros(chunk_shape, dtype=data.dtype)
        jj = 0
        for ii in indices:
            chunk[jj] = data[ii]
            if (jj + 1) % chunk_size == 0:
                jj = 0
                yield chunk
            else:
                jj += 1
        # yield remainder
        if jj:
            yield chunk[:jj]


def store_filtered_feature(rtdc_writer, feat, data, filtarr):
    """Append filtered feature data to an HDF5 file

    Parameters
    ----------
    rtdc_writer: dclab.rtdc_dataset.writer.RTDCWriter
        an open writer object
    feat: str
        feature name
    data: object or list or np.ndarray or dict
        feature data
    filtarr: boolean np.ndarray
        filtering array (same as RTDCBase.filter.all)

    Notes
    -----
    This code is somewhat redundant to the code of RTDCWriter.
    """
    indices = np.where(filtarr)[0]
    if indices.size == 0:
        warnings.warn(f"No data to export to '{rtdc_writer.path}'")
        return

    hw = rtdc_writer
    if not hw.mode == "append":
        raise ValueError("The `rtdc_writer` object must be created with"
                         + f"`mode='append'`, got '{hw.mode}' for '{hw}'!")
    # event-wise, because
    # - tdms-based datasets don't allow indexing with numpy
    # - there might be memory issues
    if feat == "contour":
        for ii in indices:
            hw.store_feature("contour", data[ii])
    elif feat in ["mask", "image", "image_bg"]:
        # assemble filtered image stacks
        for imstack in yield_filtered_array_stacks(data, indices):
            hw.store_feature(feat, imstack)
    elif feat == "trace":
        # assemble filtered trace stacks
        for tr in data.keys():
            for trstack in yield_filtered_array_stacks(data[tr], indices):
                hw.store_feature("trace", {tr: trstack})
    elif dfn.scalar_feature_exists(feat):
        hw.store_feature(feat, data[filtarr])
    else:
        # Special case of plugin or temporary features.
        shape = data[0].shape
        for dstack in yield_filtered_array_stacks(data, indices):
            hw.store_feature(feat, dstack, shape=shape)


def hdf5_append(h5obj, rtdc_ds, feat, compression, filtarr=None,
                time_offset=0):
    """Append feature data to an HDF5 file

    Parameters
    ----------
    h5obj: h5py.File
        Opened HDF5 file
    rtdc_ds: dclab.rtdc_dataset.RTDCBase
        Instance from which to obtain the data
    feat: str
        Valid feature name in `rtdc_ds`
    compression: str or None
        Compression method for "contour", "image", and "trace" data
        as well as logs; one of [None, "lzf", "gzip", "szip"].
    filtarr: None or 1d boolean np.ndarray
        Optional boolean array used for filtering. If set to
        `None`, all events are saved.
    time_offset: float
        Do not use! Please use `dclab.cli.task_join.join` instead.

    Notes
    -----
    Please update the "experiment::event count" attribute manually.
    You may use
    :func:`dclab.rtdc_dataset.writer.RTDCWriter.rectify_metadata`
    for that or use the `RTDCWriter` context manager where it is
    automatically run during `__exit__`.
    """
    # optional array for filtering events
    if filtarr is None:
        filtarr = np.ones(len(rtdc_ds), dtype=bool)
        no_filter = True
    else:
        no_filter = False

    warnings.warn("`hdf5_append` is deptecated; please use "
                  " the dclab.RTDCWriter context manager or the "
                  " export.store_filtered_feature function.",
                  DeprecationWarning)

    if time_offset != 0:
        raise ValueError("Setting `time_offset` not supported anymore! "
                         "Please use `dclab.cli.task_join.join` instead.")

    # writer instance
    hw = RTDCWriter(h5obj, mode="append", compression=compression)
    if no_filter:
        hw.store_feature(feat, rtdc_ds[feat])
    else:
        store_filtered_feature(rtdc_writer=hw,
                               feat=feat,
                               data=rtdc_ds[feat],
                               filtarr=filtarr)


def hdf5_autocomplete_config(path_or_h5obj):
    """Autocomplete the configuration of the RTDC-measurement

    The following configuration keys are updated:

    - experiment:event count
    - fluorescence:samples per event
    - imaging: roi size x (if image or mask is given)
    - imaging: roi size y (if image or mask is given)

    The following configuration keys are added if not present:

    - fluorescence:channel count

    Parameters
    ----------
    path_or_h5obj: pathlib.Path or str or h5py.File
        Path to or opened RT-DC measurement
    """
    warnings.warn("`hdf5_autocomplete_config` is deptecated; please use "
                  " the dclab.RTDCWriter context manager or the "
                  " dclab.RTDCWriter.rectify_metadata function.",
                  DeprecationWarning)
    if not isinstance(path_or_h5obj, h5py.File):
        close = True
    else:
        close = False

    hw = RTDCWriter(path_or_h5obj, mode="append")
    hw.rectify_metadata()

    if close:
        path_or_h5obj.close()



import json
from os.path import join
import pathlib
import tempfile
from unittest import mock

import h5py
import numpy as np
import pytest

import dclab
from dclab import dfn, new_dataset, RTDCWriter
from dclab.rtdc_dataset import rtdc_copy, writer
from dclab.rtdc_dataset.export import (
    ContourNotExportedWarning,
    store_filtered_feature,
)
from dclab.rtdc_dataset import check

from helper_methods import example_data_dict, retrieve_data


def test_hdf5():
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=127, keys=keys)
    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.config["imaging"]["frame rate"] = 2000

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5.rtdc")
    ds1.export.hdf5(f1, keys)

    ds2 = dclab.new_dataset(f1)
    assert ds1 != ds2
    assert np.allclose(ds2["area_um"], ds1["area_um"])
    assert np.allclose(ds2["deform"], ds1["deform"])
    assert np.allclose(ds2["time"], ds1["time"])
    assert np.allclose(ds2["frame"], ds1["frame"])
    assert np.allclose(ds2["fl3_width"], ds1["fl3_width"])


def test_hdf5_basin_ignore_internal_basin():
    """
    Internal basin definitions should not be written to exported file.
    """
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("input_file.rtdc")
    h5path_out = h5path.with_name("exported.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_copy(src_h5file=src,
                  dst_h5file=hw.h5file,
                  features="scalar")
        assert "basins" not in hw.h5file, "no basins in input file"
        # store scalar and non-scalar internal basins in the input file
        hw.store_basin(basin_name="scalar and non-scalar basin data",
                       basin_type="internal",
                       basin_format="h5dataset",
                       basin_locs=["basin_events"],
                       basin_descr="an example test basin",
                       internal_data={"userdef1": np.arange(2),
                                      "image_bg": np.zeros((2, 80, 320)),
                                      },
                       basin_map=np.zeros(src["events/deform"].shape[0]),
                       basin_feats=["image_bg", "userdef1"],
                       )

    with new_dataset(h5path_small) as ds:
        # sanity check
        assert "userdef1" in ds.features_basin
        assert "userdef1" not in ds.features_innate
        assert "image_bg" in ds.features_basin
        assert "image_bg" not in ds.features_innate
        # export that
        ds.export.hdf5(h5path_out,
                       features=["userdef1", "deform"],
                       basins=True)

    with new_dataset(h5path_out) as dso:
        assert "userdef1" in dso.features_innate
        assert "image_bg" in dso.features_basin

    with h5py.File(h5path_out) as h5:
        assert "basin_events" not in h5


def test_hdf5_basin_ignore_unknown_basin_format():
    """
    When exporting basins, unknown basin formats (those that are not in
    `get_basin_classes`) should not be exported to the output file.
    """
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("input_file.rtdc")
    h5path_out = h5path.with_name("exported.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_copy(src_h5file=src,
                  dst_h5file=hw.h5file,
                  features="scalar")
        assert "basins" not in hw.h5file, "no basins in input file"
        # store an unknown basin format in the input file
        hw.store_basin(basin_name="scalar and non-scalar basin data",
                       basin_type="file",
                       basin_format="hdf5",
                       basin_descr="a test basin with an unknown format",
                       basin_locs=[str(h5path)]
                       )

    with new_dataset(h5path_small) as ds:
        # sanity check
        assert len(ds.basins) == 1
        # modify the basins (this is a bit hacky)
        ds._basins[0].basin_format = "unknown"
        # export that
        ds.export.hdf5(h5path_out,
                       features=["deform"],
                       basins=True)

    with new_dataset(h5path_out) as dso:
        assert len(ds.basins) == 1, "because the unknown basin is ignored"
        assert "deform" in dso.features_innate
        assert "image" in dso.features_basin


def test_hdf5_basin_ignore_perishable_basins():
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("input_file.rtdc")
    h5path_out = h5path.with_name("exported.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_copy(src_h5file=src,
                  dst_h5file=hw.h5file,
                  features="scalar")
        assert "basins" not in hw.h5file, "no basins in input file"
        # store an unknown basin format in the input file
        with pytest.warns(writer.StoringPerishableBasinWarning,
                          match="perishable basin"):
            hw.store_basin(basin_name="scalar and non-scalar basin data",
                           basin_type="file",
                           basin_format="hdf5",
                           basin_descr="a test basin with an unknown format",
                           basin_locs=[str(h5path)],
                           perishable=True,
                           )

    with new_dataset(h5path_small) as ds:
        # sanity check
        assert len(ds.basins) == 1
        # Make sure there is a perishable record
        assert ds.basins[0].perishable
        # export, should ignore the perishable basin
        ds.export.hdf5(h5path_out,
                       features=["deform"],
                       basins=True)

    with new_dataset(h5path_out) as dso:
        assert len(ds.basins) == 1, "because the perishable basin is ignored"
        assert "deform" in dso.features_innate
        assert "image" in dso.features_basin


def test_hdf5_compressed():
    path = retrieve_data("fmt-hdf5_raw-cytoshot-exported.zip")
    path_exp = path.parent / "exported.rtdc"

    with dclab.new_dataset(path) as ds:
        # sanity checks
        assert len(ds.tables) == 2
        assert len(ds.logs) == 2
        ds.export.hdf5(path_exp, tables=True, logs=True)

    with dclab.new_dataset(path_exp) as ds:
        assert len(ds.tables) == 2
        assert len(ds.logs) == 3

        # check compression
        with check.IntegrityChecker(ds) as ic:
            cues = ic.check_compression()
            assert len(cues) == 1
            cue = cues[0]
            assert cue.msg == "Compression: All"


def test_hdf5_contour_ignored():
    """Contours should not be exported by default"""
    n = 65
    keys = ["contour", "image", "trace"]
    ddict = example_data_dict(size=n, keys=keys)

    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5_image.rtdc")
    with pytest.warns(ContourNotExportedWarning,
                      match="not exported to output file"):
        ds1.export.hdf5(f1, keys, filtered=False)

    # check whether contour does not exist
    with h5py.File(f1) as h5:
        assert "contour" not in h5["events"]


def test_hdf5_duplicate_feature_for_export():
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=127, keys=keys)
    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.config["imaging"]["frame rate"] = 2000

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5.rtdc")
    keys.append("area_um")
    ds1.export.hdf5(f1, keys)

    ds2 = dclab.new_dataset(f1)
    assert ds1 != ds2
    assert np.allclose(ds2["area_um"], ds1["area_um"])
    assert np.allclose(ds2["deform"], ds1["deform"])
    assert np.allclose(ds2["time"], ds1["time"])
    assert np.allclose(ds2["frame"], ds1["frame"])
    assert np.allclose(ds2["fl3_width"], ds1["fl3_width"])


def test_hdf5_contour_image_trace():
    n = 65
    keys = ["contour", "image", "trace"]
    ddict = example_data_dict(size=n, keys=keys)

    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5_image.rtdc")
    ds1.export.hdf5(f1, keys, filtered=False, allow_contour=True)
    ds2 = dclab.new_dataset(f1)

    for ii in range(n):
        assert np.all(ds1["image"][ii] == ds2["image"][ii])
        assert np.all(ds1["contour"][ii] == ds2["contour"][ii])

    for key in dfn.FLUOR_TRACES:
        assert np.all(ds1["trace"][key] == ds2["trace"][key])


def test_hdf5_contour_image_trace_large():
    """Same test for large event numbers (to test chunking)"""
    n = 653
    keys = ["contour", "image", "trace"]
    ddict = example_data_dict(size=n, keys=keys)

    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5_image.rtdc")
    ds1.export.hdf5(f1, keys, filtered=False, allow_contour=True)
    ds2 = dclab.new_dataset(f1)

    # the trace may have negative values, it's int16, not uint16.
    assert ds1["trace"]["fl1_median"][0][0] == -1, "sanity check"

    for ii in range(n):
        assert np.all(ds1["image"][ii] == ds2["image"][ii])
        assert np.all(ds1["contour"][ii] == ds2["contour"][ii])

    for key in dfn.FLUOR_TRACES:
        assert np.all(ds1["trace"][key] == ds2["trace"][key])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hdf5_contour_from_hdf5():
    ds1 = new_dataset(retrieve_data("fmt-hdf5_image-bg_2020.zip"))
    assert ds1["contour"].shape == (5, np.nan, 2)

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5_image.rtdc")
    ds1.export.hdf5(f1, ["contour"], filtered=False, allow_contour=True)
    ds2 = dclab.new_dataset(f1)
    assert ds2["contour"].shape == (5, np.nan, 2)


def test_hdf5_filtered():
    n = 10
    keys = ["area_um", "image"]
    ddict = example_data_dict(size=n, keys=keys)
    ddict["image"][3] = np.arange(10 * 20, dtype=np.uint8).reshape(10, 20) + 22

    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.filter.manual[2] = False
    ds1.apply_filter()
    fta = ds1.filter.manual.copy()

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5_filtered.rtdc")
    ds1.export.hdf5(f1, keys)

    ds2 = dclab.new_dataset(f1)

    assert ds1 != ds2
    assert np.allclose(ds2["area_um"], ds1["area_um"][fta])
    assert np.allclose(ds2["image"][2], ds1["image"][3])
    assert np.all(ds2["image"][2] != ds1["image"][2])


def test_hdf5_filtered_empty():
    n = 10
    keys = ["area_um", "image"]
    ddict = example_data_dict(size=n, keys=keys)

    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    # Remove all events
    ds1.filter.manual[:] = False
    ds1.apply_filter()

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5_filtered.rtdc")
    with pytest.warns(UserWarning, match="No data to export to"):
        ds1.export.hdf5(f1, keys)

    with h5py.File(f1) as h5:
        assert len(h5["events"].keys()) == 0


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.parametrize("dataset", ["fmt-hdf5_mask-contour_2018.zip",
                                     "fmt-hdf5_fl-no-contour_2019.zip"])
def test_hdf5_filtered_full(dataset):
    """Compare full export to filter with all-True"""
    tdir = pathlib.Path(tempfile.mkdtemp())
    path_in = tdir / "in.rtdc"

    # first, create a large dataset
    with new_dataset(retrieve_data(dataset)) as ds0:
        num = 100 % len(ds0) + len(ds0)
        ds0.export.hdf5(path_in,
                        features=ds0.features_innate,
                        allow_contour=True)
        with RTDCWriter(path_in) as hw:
            for _ in range(num):
                for feat in ds0.features_innate:
                    hw.store_feature(feat, ds0[feat])

    # 1. via export unfiltered
    path_out1 = tdir / "out1.rtdc"
    with dclab.new_dataset(path_in) as ds:
        assert len(ds) >= 100
        ds.export.hdf5(path_out1,
                       features=ds.features_innate,
                       filtered=False,
                       allow_contour=True)

    # 2. via export filtered
    path_out2 = tdir / "out2.rtdc"
    with dclab.new_dataset(path_in) as ds:
        ds.export.hdf5(path_out2,
                       features=ds.features_innate,
                       filtered=True,
                       allow_contour=True)

    # 3. via RTDC writer
    path_out3 = tdir / "out3.rtdc"
    with dclab.new_dataset(path_in) as ds:
        with RTDCWriter(path_out3) as hw:
            config = ds.config.as_dict(pop_filtering=True)
            hw.store_metadata(config)
            for feat in ds.features_innate:
                hw.store_feature(feat, ds[feat])

    # 4. via export filtered function
    path_out4 = tdir / "out4.rtdc"
    with dclab.new_dataset(path_in) as ds:
        with RTDCWriter(path_out4) as hw:
            config = ds.config.as_dict(pop_filtering=True)
            hw.store_metadata(config)
            for feat in ds.features_innate:
                store_filtered_feature(rtdc_writer=hw,
                                       feat=feat,
                                       data=ds[feat],
                                       filtarr=np.ones(len(ds), dtype=bool))

    with dclab.new_dataset(path_in) as ds0:
        for path in [path_out1, path_out2, path_out3, path_out4]:
            with dclab.new_dataset(path) as dsi:
                for feat in dsi.features_innate:
                    if dfn.scalar_feature_exists(feat):
                        assert np.all(ds0[feat] == dsi[feat])
                    elif feat == "trace":
                        for key in ds0[feat]:
                            assert np.all(np.array(ds0[feat][key])
                                          == np.array(dsi[feat][key]))
                    else:
                        for ii in range(len(ds0)):
                            assert np.all(ds0[feat][ii] == dsi[feat][ii])


def test_hdf5_filtered_index():
    """Make sure that exported index is always re-enumerated"""
    n = 10
    keys = ["area_um", "deform", "index"]
    ddict = example_data_dict(size=n, keys=keys)
    ddict["index"] = np.arange(1, n+1)

    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.filter.manual[2] = False
    ds1.apply_filter()

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5_filtered.rtdc")
    ds1.export.hdf5(f1, keys)

    ds2 = dclab.new_dataset(f1)

    assert len(ds2) == n - 1
    assert np.all(ds2["index"] == np.arange(1, n))
    assert ds2.config["experiment"]["event count"] == n - 1


@pytest.mark.parametrize("export_filtering", [True, False])
def test_hdf5_hierarchy_basin_only_export(export_filtering):
    """
    When the dataset that is exported is a hierarchy child,
    then there should be a mapped basin in the output file
    referring to the input data.

    The parametrization of this method checks whether the output
    is the same when `filtered` is set or not. It should be the
    same, because the hierarchy child has no filters defined.
    """
    h5path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    path_exp = h5path.with_name("exported.rtdc")

    with dclab.new_dataset(h5path) as ds:
        ds.filter.manual[:5] = False
        ds.apply_filter()

        with dclab.new_dataset(ds) as dsc:
            dsc.export.hdf5(path_exp,
                            features=[],
                            basins=True,
                            filtered=export_filtering,
                            )

    with h5py.File(path_exp) as h5:
        assert "basinmap0" in h5["events"]
        assert "basins" in h5
        assert len(h5["basins"]) == 1

    with dclab.new_dataset(path_exp) as dse, dclab.new_dataset(h5path) as ds:
        assert "image" in dse.features_basin
        assert "mask" in dse.features_basin
        assert "area_um" in dse.features_basin

        assert "image" not in dse.features_innate
        assert "mask" not in dse.features_innate
        assert "area_um" not in dse.features_innate

        # check for filtering
        assert np.allclose(dse["deform"][2], 0.14734814,
                           atol=1e-7, rtol=0)
        assert np.all(dse["deform"][:] == ds["deform"][5:])


def test_hdf5_hierarchy_basin_only_export_with_filters():
    """
    When the dataset that is exported is a hierarchy child,
    then there should be a mapped basin in the output file
    referring to the input data.

    This test performs the export after an additional filter
    is applied to the hierarchy dataset.
    """
    h5path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    path_exp = h5path.with_name("exported.rtdc")

    with dclab.new_dataset(h5path) as ds:
        ds.filter.manual[:5] = False
        ds.apply_filter()

        with dclab.new_dataset(ds) as dsc:
            dsc.filter.manual[::2] = False
            dsc.apply_filter()
            dsc.export.hdf5(path_exp,
                            filtered=True,
                            features=[],
                            basins=True,
                            )

    with h5py.File(path_exp) as h5:
        assert "basinmap0" in h5["events"]
        assert "basins" in h5
        assert len(h5["basins"]) == 1

    with dclab.new_dataset(path_exp) as dse, dclab.new_dataset(h5path) as ds:
        assert "image" in dse.features_basin
        assert "mask" in dse.features_basin
        assert "area_um" in dse.features_basin

        assert "image" not in dse.features_innate
        assert "mask" not in dse.features_innate
        assert "area_um" not in dse.features_innate

        # check for inception filtering
        assert np.allclose(dse["deform"][0], 0.15025392,
                           atol=1e-7, rtol=0)
        assert np.all(dse["deform"][:] == ds["deform"][6::2])
        assert np.all(dse["deform"][:] == ds["deform"][5:][1::2]), "sanity"


def test_hdf5_hierarchy_basin_export_filter_hierarchy_depth_two():
    """
    When the dataset that is exported is a hierarchy child,
    then there should be a mapped basin in the output file
    referring to the input data.

    This test performs the export after two hierarchy children are
    created with an additional filter applied to the last hierarchy dataset.
    """
    h5path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    path_exp = h5path.with_name("exported.rtdc")

    with dclab.new_dataset(h5path) as ds:
        ds.filter.manual[:5] = False
        ds.apply_filter()

        with dclab.new_dataset(ds) as dsc:
            dsc.filter.manual[::2] = False
            dsc.apply_filter()
            with dclab.new_dataset(dsc) as dsci:
                dsci.filter.manual[-1] = False
                dsci.apply_filter()
                dsci.export.hdf5(path_exp,
                                 filtered=True,
                                 features=[],
                                 basins=True,
                                 )

    with h5py.File(path_exp) as h5:
        assert "basinmap0" in h5["events"]
        assert "basins" in h5
        assert len(h5["basins"]) == 1

    with dclab.new_dataset(path_exp) as dse, dclab.new_dataset(h5path) as ds:
        assert "image" in dse.features_basin
        assert "mask" in dse.features_basin
        assert "area_um" in dse.features_basin

        assert "image" not in dse.features_innate
        assert "mask" not in dse.features_innate
        assert "area_um" not in dse.features_innate

        # check for inception filtering
        assert np.allclose(dse["deform"][0], 0.15025392,
                           atol=1e-7, rtol=0)
        assert np.all(dse["deform"][:] == ds["deform"][6::2][:-1])
        assert np.all(dse["deform"][:] == ds["deform"][5:][1::2][:-1]), "sane"


@pytest.mark.parametrize("export_filtering", [True, False])
def test_hdf5_hierarchy_basin_export_filter_hierarchy_depth_two_no_filter(
        export_filtering):
    """
    When the dataset that is exported is a hierarchy child,
    then there should be a mapped basin in the output file
    referring to the input data.

    This test performs the export after two hierarchy children are
    created without a final filter in the last hierarchy dataset.

    The parametrization of this method checks whether the output
    is the same when `filtered` is set or not. It should be the
    same, because the hierarchy child has no filters defined.
    """
    h5path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    path_exp = h5path.with_name("exported.rtdc")

    with dclab.new_dataset(h5path) as ds:
        ds.filter.manual[:5] = False
        ds.apply_filter()

        with dclab.new_dataset(ds) as dsc:
            dsc.filter.manual[::2] = False
            dsc.apply_filter()
            with dclab.new_dataset(dsc) as dsci:
                dsci.export.hdf5(path_exp,
                                 filtered=export_filtering,
                                 features=[],
                                 basins=True,
                                 )

    with h5py.File(path_exp) as h5:
        assert "basinmap0" in h5["events"]
        assert "basins" in h5
        assert len(h5["basins"]) == 1

    with dclab.new_dataset(path_exp) as dse, dclab.new_dataset(h5path) as ds:
        assert "image" in dse.features_basin
        assert "mask" in dse.features_basin
        assert "area_um" in dse.features_basin

        assert "image" not in dse.features_innate
        assert "mask" not in dse.features_innate
        assert "area_um" not in dse.features_innate

        # check for inception filtering
        assert np.allclose(dse["deform"][0], 0.15025392,
                           atol=1e-7, rtol=0)
        assert np.all(dse["deform"][:] == ds["deform"][6::2])
        assert np.all(dse["deform"][:] == ds["deform"][5:][1::2]), "sane"


def test_hdf5_hierarchy_basin_export_filter_basins_depth_two():
    """
    When the dataset that is exported is a hierarchy child,
    then there should be a mapped basin in the output file
    referring to the input data.

    This test makes sure this works if you do this twice,
    opening the exported file with dclab, filtering, and exporting again.
    """
    h5path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    path_exp1 = h5path.with_name("exported-1.rtdc")
    path_exp2 = h5path.with_name("exported-2.rtdc")

    with dclab.new_dataset(h5path) as ds:
        ds.filter.manual[:5] = False
        ds.apply_filter()
        ds.export.hdf5(path_exp1,
                       filtered=True,
                       features=[],
                       basins=True,
                       )

    with h5py.File(path_exp1) as h5:
        assert "basinmap0" in h5["events"]
        assert "basins" in h5
        assert len(h5["basins"]) == 1

    with dclab.new_dataset(path_exp1) as dsc:
        dsc.filter.manual[::2] = False
        dsc.apply_filter()
        with dclab.new_dataset(dsc) as dsci:
            dsci.filter.manual[-1] = False
            dsci.apply_filter()
            dsci.export.hdf5(path_exp2,
                             filtered=True,
                             features=[],
                             basins=True,
                             )

    with h5py.File(path_exp2) as h5:
        assert "basinmap1" in h5["events"]
        assert "basins" in h5
        assert len(h5["basins"]) == 2

    with dclab.new_dataset(path_exp2) as dse, dclab.new_dataset(h5path) as ds:
        assert "image" in dse.features_basin
        assert "mask" in dse.features_basin
        assert "area_um" in dse.features_basin

        assert "image" not in dse.features_innate
        assert "mask" not in dse.features_innate
        assert "area_um" not in dse.features_innate

        # check for inception filtering
        assert np.allclose(dse["deform"][0], 0.15025392,
                           atol=1e-7, rtol=0)
        assert np.all(dse["deform"][:] == ds["deform"][6::2][:-1])
        assert np.all(dse["deform"][:] == ds["deform"][5:][1::2][:-1]), "sane"


@pytest.mark.parametrize("export_filtering", [True, False])
def test_hdf5_hierarchy_basin_export_filter_basins_depth_two_no_filter(
        export_filtering):
    """
    When the dataset that is exported is a hierarchy child,
    then there should be a mapped basin in the output file
    referring to the input data.

    This test makes sure this works if you do this twice,
    opening the exported file with dclab, filtering, and exporting again.

    The parametrization of this method checks whether the output
    is the same when `filtered` is set or not. It should be the
    same, because the hierarchy child has no filters defined.
    """
    h5path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    path_exp1 = h5path.with_name("exported-1.rtdc")
    path_exp2 = h5path.with_name("exported-2.rtdc")

    with dclab.new_dataset(h5path) as ds:
        ds.filter.manual[:5] = False
        ds.apply_filter()
        ds.export.hdf5(path_exp1,
                       filtered=True,
                       features=[],
                       basins=True,
                       )

    with h5py.File(path_exp1) as h5:
        assert "basinmap0" in h5["events"]
        assert "basins" in h5
        assert len(h5["basins"]) == 1

    with dclab.new_dataset(path_exp1) as dsc:
        dsc.filter.manual[::2] = False
        dsc.apply_filter()
        with dclab.new_dataset(dsc) as dsci:
            dsci.apply_filter()
            dsci.export.hdf5(path_exp2,
                             filtered=export_filtering,
                             features=[],
                             basins=True,
                             )

    with h5py.File(path_exp2) as h5:
        assert "basinmap1" in h5["events"]
        assert "basins" in h5
        assert len(h5["basins"]) == 2

    with dclab.new_dataset(path_exp2) as dse, dclab.new_dataset(h5path) as ds:
        assert "image" in dse.features_basin
        assert "mask" in dse.features_basin
        assert "area_um" in dse.features_basin

        assert "image" not in dse.features_innate
        assert "mask" not in dse.features_innate
        assert "area_um" not in dse.features_innate

        # check for inception filtering
        assert np.allclose(dse["deform"][0], 0.15025392,
                           atol=1e-7, rtol=0)
        assert np.all(dse["deform"][:] == ds["deform"][6::2])
        assert np.all(dse["deform"][:] == ds["deform"][5:][1::2]), "sane"


@pytest.mark.parametrize("export_filtering", [True, False])
def test_hdf5_normal_basin_export_filter_basins_depth_two_no_filter(
        export_filtering):
    """
    Filter and export a dataset, then open and export that dataset unfiltered.

    The parametrization of this method checks whether the output
    is the same when `filtered` is set or not. It should be the
    same, because the hierarchy child has no filters defined.
    """
    h5path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    path_exp1 = h5path.with_name("exported-1.rtdc")
    path_exp2 = h5path.with_name("exported-2.rtdc")

    with dclab.new_dataset(h5path) as ds:
        ds.filter.manual[:5] = False
        ds.apply_filter()
        ds.export.hdf5(path_exp1,
                       filtered=True,
                       features=[],
                       basins=True,
                       )

    with h5py.File(path_exp1) as h5:
        assert "basinmap0" in h5["events"]
        assert "basins" in h5
        assert len(h5["basins"]) == 1

    with dclab.new_dataset(path_exp1) as dsc:
        dsc.apply_filter()
        dsc.export.hdf5(path_exp2,
                        filtered=export_filtering,
                        features=[],
                        basins=True,
                        )

    with h5py.File(path_exp2) as h5:
        assert "basinmap0" in h5["events"]
        if export_filtering:
            assert "basinmap1" in h5["events"]
        else:
            assert "basinmap1" not in h5["events"]
        assert "basins" in h5
        assert len(h5["basins"]) == 2

    with dclab.new_dataset(path_exp2) as dse, dclab.new_dataset(h5path) as ds:
        assert "image" in dse.features_basin
        assert "mask" in dse.features_basin
        assert "area_um" in dse.features_basin

        assert "image" not in dse.features_innate
        assert "mask" not in dse.features_innate
        assert "area_um" not in dse.features_innate

        # check for inception filtering
        assert np.allclose(dse["deform"][1], 0.15025392,
                           atol=1e-7, rtol=0)
        assert np.all(dse["deform"][:] == ds["deform"][5:]), "sane"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hdf5_identifier_different():
    """Make sure the identifier is changed upon export"""
    pin = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    pout = pin.with_name("exported.rtdc")
    with new_dataset(pin) as din:
        mid_in = din.get_measurement_identifier()
        assert len(din) == 5
        assert mid_in
        din.config["filtering"]["deform min"] = np.mean(din["deform"])
        din.config["filtering"]["deform max"] = np.max(din["deform"])
        din.apply_filter()
        din.export.hdf5(pout, ["area_um", "deform"], filtered=True)

    with new_dataset(pout) as dout:
        mid_out = dout.get_measurement_identifier()
        assert mid_out.startswith(mid_in)
        assert mid_out != mid_in
        assert len(dout) == 3


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hdf5_identifier_same_unfiltered():
    """Make sure the identifier is changed upon export"""
    pin = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    pout = pin.with_name("exported.rtdc")
    with new_dataset(pin) as din:
        mid_in = din.get_measurement_identifier()
        assert len(din) == 5
        assert mid_in
        din.config["filtering"]["deform min"] = np.mean(din["deform"])
        din.config["filtering"]["deform max"] = np.max(din["deform"])
        din.apply_filter()
        din.export.hdf5(pout, ["area_um", "deform"], filtered=False)

    with new_dataset(pout) as dout:
        mid_out = dout.get_measurement_identifier()
        assert mid_out == mid_in
        assert len(dout) == 5


def test_hdf5_image_bg():
    n = 65
    keys = ["image", "image_bg"]
    ddict = example_data_dict(size=n, keys=keys)

    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5_image_bg.rtdc")
    ds1.export.hdf5(f1, keys, filtered=False)
    ds2 = dclab.new_dataset(f1)

    for ii in range(n):
        assert np.all(ds1["image"][ii] == ds2["image"][ii])
        assert np.all(ds1["image_bg"][ii] == ds2["image_bg"][ii])


def test_hdf5_index_continuous():
    keys = ["area_um", "deform", "time", "frame", "index"]
    ddict = example_data_dict(size=10, keys=keys)
    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.config["imaging"]["frame rate"] = 2000

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5.rtdc")
    ds1.export.hdf5(f1, keys)
    with RTDCWriter(f1) as hw:
        for feat in keys:
            hw.store_feature(feat, ds1[feat])

    with h5py.File(f1, "r") as h5:
        assert "index" in h5["events"]
        assert np.allclose(h5["events/index"][:], np.arange(1, 21))


def test_hdf5_index_online_continuous():
    keys = ["area_um", "deform", "time", "frame", "index_online"]
    ddict = example_data_dict(size=10, keys=keys)
    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.config["imaging"]["frame rate"] = 2000

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5.rtdc")
    ds1.export.hdf5(f1, keys)
    with RTDCWriter(f1) as hw:
        for feat in keys:
            hw.store_feature(feat, ds1[feat])

    with h5py.File(f1, "r") as h5:
        assert "index_online" in h5["events"]


def test_hdf5_index_online_replaces_index():
    keys = ["area_um", "deform", "time", "frame", "index_online"]
    ddict = example_data_dict(size=10, keys=keys)
    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.config["imaging"]["frame rate"] = 2000

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5.rtdc")
    ds1.export.hdf5(f1, keys)
    with h5py.File(f1, "r") as h5:
        assert "index" not in h5["events"]
        assert "index_online" in h5["events"]


@pytest.mark.parametrize("logs", [True, False])
def test_hdf5_logs(logs):
    keys = ["area_um", "deform", "time", "frame", "index_online"]
    ddict = example_data_dict(size=10, keys=keys)
    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.config["imaging"]["frame rate"] = 2000
    ds1.logs["iratrax"] = ["don't", "eat", "brown", "apples"]

    edest = pathlib.Path(tempfile.mkdtemp())
    f1 = edest / "dclab_test_export_hdf5_1.rtdc"
    ds1.export.hdf5(f1, keys, logs=logs)
    with h5py.File(f1, "r") as h5:
        if logs:
            assert "src_iratrax" in h5.get("logs", {})
        else:
            assert "src_iratrax" not in h5.get("logs", {})


def test_hdf5_logs_meta():
    """Test export of export log #251"""
    keys = ["area_um", "deform", "fl3_width", "frame", "time"]
    ddict = example_data_dict(size=127, keys=keys)
    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.config["imaging"]["frame rate"] = 2000

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5.rtdc")
    ds1.export.hdf5(f1, keys)

    ds2 = dclab.new_dataset(f1)
    for name in ds2.logs:
        if name.startswith("dclab-export_"):
            kwargs = json.loads("\n".join(ds2.logs[name]))["kwargs"]
            assert kwargs["features"] == keys
            break
    else:
        assert False


def test_hdf5_ml_score():
    data = {"ml_score_ds9": [.80, .31, .12, .01, .59, .40, .52],
            "ml_score_dsc": [.12, .52, .21, .24, .30, .22, .79],
            "ml_score_ent": [.42, .11, .78, .11, .54, .24, .15],
            "ml_score_pic": [.30, .30, .10, .99, .59, .55, .19],
            "ml_score_tng": [.13, .33, .13, .01, .79, .11, .22],
            "ml_score_tos": [.14, .34, .12, .01, .59, .56, .17],
            "ml_score_voy": [.25, .12, .42, .33, .21, .55, .82],
            }
    ds0 = dclab.new_dataset(data)
    ds0.config["experiment"]["sample"] = "test"
    ds0.config["experiment"]["run index"] = 1

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.rtdc")
    ds0.export.hdf5(f1, list(data.keys()), override=True)

    with dclab.new_dataset(f1) as ds:
        assert np.allclose(ds["ml_class"], np.arange(7))
        assert "ml_class" not in ds._events
        assert "ml_class" in ds._ancillaries


def test_hdf5_override():
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=212, keys=keys)
    ds = dclab.new_dataset(ddict)
    ds.config["imaging"]["frame rate"] = 2000

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.rtdc")
    ds.export.hdf5(f1, keys, override=True)
    try:
        ds.export.hdf5(f1[:-5], keys, override=False)
    except OSError:
        pass
    else:
        raise ValueError("Should append .rtdc and not override!")


def test_hdf5_progress(tmp_path):
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=222, keys=keys)
    ds = dclab.new_dataset(ddict)

    f1 = tmp_path / "test.rtdc"

    callback = mock.MagicMock()
    ds.export.hdf5(f1, keys, progress_callback=callback)

    callback.assert_any_call(0.0, "writing metadata")
    callback.assert_any_call(0/5, "exporting area_um")
    callback.assert_any_call(1/5, "exporting deform")
    callback.assert_any_call(2/5, "exporting fl3_width")
    callback.assert_any_call(3/5, "exporting frame")
    callback.assert_any_call(4/5, "exporting time")
    callback.assert_any_call(1.0, "export complete")


def test_hdf5_progress_basins(tmp_path):
    h5path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    path_exp = h5path.with_name("exported.rtdc")

    callback = mock.MagicMock()

    with dclab.new_dataset(h5path) as ds:
        ds.export.hdf5(path_exp,
                       features=["deform", "image"],
                       basins=True,
                       progress_callback=callback
                       )

    callback.assert_any_call(0.0, "writing metadata")
    callback.assert_any_call(0/2, "exporting deform")
    callback.assert_any_call(1/2, "exporting image")
    callback.assert_any_call(1/2, "writing basins")
    callback.assert_any_call(1.0, "export complete")


def test_hdf5_progress_basins_only(tmp_path):
    h5path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    path_exp = h5path.with_name("exported.rtdc")

    callback = mock.MagicMock()

    with dclab.new_dataset(h5path) as ds:
        ds.export.hdf5(path_exp,
                       features=[],
                       basins=True,
                       progress_callback=callback
                       )

    callback.assert_any_call(0.0, "writing metadata")
    callback.assert_any_call(0.0, "writing basins")
    callback.assert_any_call(1.0, "export complete")


@pytest.mark.parametrize("tables", [True, False])
def test_hdf5_tables(tables):
    keys = ["area_um", "deform", "time", "frame", "index_online"]
    ddict = example_data_dict(size=10, keys=keys)
    ds1 = new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.config["imaging"]["frame rate"] = 2000

    # generate a table
    columns = ["bread", "beer", "chocolate"]
    ds_dt = np.dtype({'names': columns,
                      'formats': [np.float64] * len(columns)})
    tab_data = np.zeros((10, len(columns)))
    tab_data[:, 0] = np.arange(10)
    tab_data[:, 1] = 1000
    tab_data[:, 2] = np.linspace(np.pi, 2*np.pi, 10)
    rec_arr = np.rec.array(tab_data, dtype=ds_dt)

    ds1.tables["iratrax"] = rec_arr

    edest = pathlib.Path(tempfile.mkdtemp())
    f1 = edest / "dclab_test_export_hdf5_1.rtdc"
    ds1.export.hdf5(f1, keys, tables=tables)
    with h5py.File(f1, "r") as h5:
        if tables:
            assert "src_iratrax" in h5.get("tables", {})
        else:
            assert "src_iratrax" not in h5.get("tables", {})

    if tables:
        # reading the file with dclab
        with new_dataset(f1) as ds:
            assert "src_iratrax" in ds.tables
            assert ds.tables["src_iratrax"].has_graphs()
            assert np.all(ds.tables["src_iratrax"] == rec_arr)


def test_hdf5_tables_array_only():
    keys = ["area_um", "deform", "time", "frame", "index_online"]
    ddict = example_data_dict(size=10, keys=keys)
    ds1 = new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.config["imaging"]["frame rate"] = 2000

    # generate a table that consists of an array, not a dict-like object
    tab_data = np.random.random((1000, 300))
    ds1.tables["iratrax"] = tab_data

    edest = pathlib.Path(tempfile.mkdtemp())
    f1 = edest / "dclab_test_export_hdf5_1.rtdc"
    ds1.export.hdf5(f1, keys, tables=True)
    with h5py.File(f1, "r") as h5:
        assert "src_iratrax" in h5.get("tables", {})
        assert np.all(h5["tables"]["src_iratrax"] == tab_data)
        assert h5["tables"]["src_iratrax"].attrs["CLASS"] == np.bytes_("IMAGE")

    # reading the file with dclab
    with new_dataset(f1) as ds:
        assert "src_iratrax" in ds.tables
        assert not ds.tables["src_iratrax"].has_graphs()
        assert np.all(ds.tables["src_iratrax"] == tab_data)
        assert ds.tables["src_iratrax"].meta["CLASS"] == np.bytes_("IMAGE")


def test_hdf5_table_from_file(tmp_path):
    ds = new_dataset(retrieve_data("fmt-hdf5_raw-cytoshot-exported.zip"))

    f1 = join(tmp_path, "test.rtdc")
    ds.export.hdf5(f1, ["temp"], tables=True)

    ds2 = new_dataset(f1)
    tab = ds2.tables["src_so2exp_cytoshot_monitor"]

    assert np.allclose(
        tab["focus"][0],
        6.241900073522602,
        atol=1e-7, rtol=0)

    assert tab.meta["COLOR_focus"] == "#a03000"


def test_hdf5_table_from_file_with_unnamed_table(tmp_path):
    path = retrieve_data("fmt-hdf5_raw-cytoshot-exported.zip")
    with RTDCWriter(path) as hw:
        hw.store_table("just_an_array", np.ones((30, 50)),
                       h5_attrs={"heinz": "kunz"})

    # Make sure that worked with tables
    with new_dataset(path) as ds:
        tab = ds.tables["just_an_array"]
        assert not tab.has_graphs()
        assert np.array(tab).shape == (30, 50)
        assert tab.meta["heinz"] == "kunz"

    with new_dataset(path) as ds:
        f1 = join(tmp_path, "test.rtdc")
        ds.export.hdf5(f1, ["temp"], tables=True)

    with new_dataset(f1) as ds2:
        tab = ds2.tables["src_so2exp_cytoshot_monitor"]
        assert tab.has_graphs()
        assert np.allclose(
            tab["focus"][0],
            6.241900073522602,
            atol=1e-7, rtol=0)

        assert tab.meta["COLOR_focus"] == "#a03000"

        tab2 = ds2.tables["src_just_an_array"]
        assert not tab2.has_graphs()
        assert np.array(tab2).shape == (30, 50)
        assert tab2.meta["heinz"] == "kunz"


def test_hdf5_trace_from_tdms():
    pytest.importorskip("nptdms")
    ds = new_dataset(retrieve_data("fmt-tdms_2fl-no-image_2017.zip"))

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.rtdc")
    ds.export.hdf5(f1, ["trace"])

    ds2 = new_dataset(f1)

    for key in ds["trace"]:
        for ii in range(len(ds)):
            assert np.all(ds["trace"][key][ii] == ds2["trace"][key][ii])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hdf5_traces():
    """Length of traces is preserved (no filtering)"""
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))

    # sanity check
    assert len(ds) == 7
    assert len(ds["trace"]["fl1_median"]) == 7

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.rtdc")
    ds.export.hdf5(f1, ["deform", "trace"])

    ds2 = new_dataset(f1)
    assert len(ds2) == 7
    assert len(ds2["trace"]["fl1_median"]) == 7


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hdf5_traces_filter():
    """Length of traces was wrong when filters were applied #112

    Test dataset length with additional feature.
    """
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))

    # applying some filters
    ds.config["filtering"]["deform min"] = 0.01
    ds.config["filtering"]["deform max"] = 0.1
    ds.apply_filter()

    # sanity check
    assert np.sum(ds.filter.all) == 3

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.rtdc")
    ds.export.hdf5(f1, ["deform", "trace"])

    ds2 = new_dataset(f1)
    assert len(ds2) == 3
    assert len(ds2["deform"]) == 3
    assert len(ds2["trace"]["fl1_median"]) == 3
    assert np.all(ds["trace"]["fl1_raw"][3] == ds2["trace"]["fl1_raw"][0])
    assert np.all(ds["trace"]["fl1_raw"][5] == ds2["trace"]["fl1_raw"][1])
    assert np.all(ds["trace"]["fl1_raw"][6] == ds2["trace"]["fl1_raw"][2])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hdf5_traces_filter2():
    """Length of traces was wrong when filters were applied #112

    Test dataset lenght only with trace.
    """
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))

    # applying some filters
    ds.config["filtering"]["deform min"] = 0.01
    ds.config["filtering"]["deform max"] = 0.1
    ds.apply_filter()

    # sanity check
    assert np.sum(ds.filter.all) == 3

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.rtdc")
    ds.export.hdf5(f1, ["deform", "trace"])

    ds2 = new_dataset(f1)
    assert len(ds2) == 3
    assert len(ds2["deform"]) == 3
    assert len(ds2["trace"]["fl1_median"]) == 3
    assert np.all(ds["trace"]["fl1_raw"][3] == ds2["trace"]["fl1_raw"][0])
    assert np.all(ds["trace"]["fl1_raw"][5] == ds2["trace"]["fl1_raw"][1])
    assert np.all(ds["trace"]["fl1_raw"][6] == ds2["trace"]["fl1_raw"][2])



import warnings

from .kde.methods import (  # noqa: F401
    bin_num_doane, bin_width_doane, bin_width_percentile, get_bad_vals,
    ignore_nan_inf, kde_gauss, kde_histogram, kde_multivariate, kde_none,
    methods
)

warnings.warn("`dclab.kde_methods` is deprecated; please use "
              "the `dclab.kde.methods` instead",
              DeprecationWarning)



from unittest import mock

import pytest

import dclab
from dclab import new_dataset, RTDCWriter
import numpy as np

from helper_methods import example_data_dict, retrieve_data

av = pytest.importorskip("av")


def test_avi_export_check(tmp_path):
    ds = new_dataset(retrieve_data("fmt-hdf5_wide-channel_2023.zip"))
    avi_path = tmp_path / "exported.avi"
    ds.export.avi(path=avi_path)
    num_frames = 0
    with av.open(avi_path) as container:
        for ii, frame in enumerate(container.decode(video=0)):
            num_frames += 1
            array = frame.to_ndarray(format="rgb24")
            for jj in range(3):
                assert np.allclose(ds["image"][ii],
                                   array[:, :, jj],
                                   atol=5,  # one reason why we use HDF5
                                   rtol=0)
    assert num_frames == len(ds)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_avi_export_progress(tmp_path):
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    # create an .rtdc file with more than 10_000 events
    path_large = path.with_name("large.rtdc")
    path_avi = path.with_name("export.avi")

    with RTDCWriter(path_large) as hw, new_dataset(path) as ds:
        hw.store_metadata(ds.config.as_dict(pop_filtering=True))
        size = len(ds)
        num_iters = 25_000 // size
        for feat in ["time", "image", "deform"]:
            feat_data = np.concatenate([ds[feat][:]] * num_iters)
            hw.store_feature(feat, feat_data)

    with dclab.new_dataset(path_large) as dse:
        assert len(dse) > 20_000, "we need a large dataset"

        callback = mock.MagicMock()
        dse.export.avi(path=path_avi, progress_callback=callback)

    callback.assert_any_call(0.0, "exporting video")
    callback.assert_any_call(0.4, "exporting video")
    callback.assert_any_call(0.8, "exporting video")
    callback.assert_any_call(1.0, "video export complete")


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'fmt_tdms.exc.CorruptFrameWarning')
def test_avi_export_tdms(tmp_path):
    pytest.importorskip("nptdms")
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image_2016.zip"))

    f1 = tmp_path / "test.avi"
    ds.export.avi(path=f1)
    assert f1.stat()[6] > 1e4, "Resulting file to small, Something went wrong!"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'fmt_tdms.exc.CorruptFrameWarning')
def test_avi_override_tdms(tmp_path):
    pytest.importorskip("nptdms")
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image_2016.zip"))

    f1 = tmp_path / "test.avi"
    ds.export.avi(f1, override=True)
    try:
        ds.export.avi(f1.with_name(f1.stem), override=False)
    except OSError:
        pass
    else:
        raise ValueError("Should append .avi and not override!")


def test_avi_no_images(tmp_path):
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=127, keys=keys)
    ds = dclab.new_dataset(ddict)

    try:
        ds.export.avi(tmp_path / "test.avi")
    except OSError:
        pass
    else:
        raise ValueError("There should be no image data to write!")



import numpy as np
import pytest
import dclab
from dclab.external import skimage
from dclab.kde import KernelDensityEstimator
from dclab.kde.methods import KernelDensityEstimationForEmtpyArrayWarning
from dclab import polygon_filter

from helper_methods import example_data_dict


def test_contour_lines():
    np.random.seed(47)
    x0 = np.random.normal(loc=100, scale=10, size=100)
    y0 = np.random.normal(loc=.1, scale=.01, size=100)

    ds = dclab.new_dataset({"area_um": x0, "deform": y0})
    ds.config["filtering"]["enable filters"] = False

    kde_instance = KernelDensityEstimator(ds)

    contours = kde_instance.get_contour_lines(xax="area_um",
                                              yax="deform",
                                              xacc=.10,
                                              yacc=.01,
                                              kde_type="histogram",
                                              kde_kwargs=None,
                                              xscale="linear",
                                              yscale="linear",
                                              quantiles=[0.89])
    nump = 0
    for p in zip(x0, y0):
        nump += polygon_filter.PolygonFilter.point_in_poly(p,
                                                           poly=contours[0][0])

    assert nump == 11, "there should be (1-q)*100 points in the contour"

    # added in dclab 0.24.1
    nump2 = skimage.measure.points_in_poly(
        np.concatenate((x0.reshape(-1, 1), y0.reshape(-1, 1)), axis=1),
        contours[0][0])
    assert nump2.sum() == 11


def test_contour_lines_without_quantiles_with_ret_levels():
    np.random.seed(47)
    x0 = np.random.normal(loc=100, scale=10, size=100)
    y0 = np.random.normal(loc=.1, scale=.01, size=100)

    ds = dclab.new_dataset({"area_um": x0, "deform": y0})
    ds.config["filtering"]["enable filters"] = False

    kde_instance = KernelDensityEstimator(ds)

    contours, levels = kde_instance.get_contour_lines(xax="area_um",
                                                      yax="deform",
                                                      xacc=.10,
                                                      yacc=.01,
                                                      kde_type="histogram",
                                                      kde_kwargs=None,
                                                      xscale="linear",
                                                      yscale="linear",
                                                      ret_levels=True,)
    nump = 0
    for p in zip(x0, y0):
        nump += polygon_filter.PolygonFilter.point_in_poly(p,
                                                           poly=contours[0][1])

    assert nump == 50, "there should be (1-q)*100 points in the contour"

    # added in dclab 0.24.1
    nump2 = skimage.measure.points_in_poly(
        np.concatenate((x0.reshape(-1, 1), y0.reshape(-1, 1)), axis=1),
        contours[0][1])
    assert nump2.sum() == 50

    assert len(levels) == 2, "there should be two levels"


@pytest.mark.filterwarnings("ignore::UserWarning", "ignore::RuntimeWarning")
def test_contour_lines_data_with_too_much_space():
    np.random.seed(47)
    x0 = np.random.normal(loc=100, scale=10, size=100)
    y0 = np.random.normal(loc=.1, scale=.01, size=100)

    n = 100  # Adjust this to control spacing
    x0 = x0[::n]
    y0 = y0[::n]

    ds = dclab.new_dataset({"area_um": x0, "deform": y0})
    ds.config["filtering"]["enable filters"] = False

    kde_instance = KernelDensityEstimator(ds)

    contours = kde_instance.get_contour_lines(xax="area_um",
                                              yax="deform",
                                              xacc=.10,
                                              yacc=.01,
                                              kde_type="histogram",
                                              kde_kwargs=None,
                                              xscale="linear",
                                              yscale="linear",
                                              quantiles=[0.5])
    assert not contours, "there should be no contours with too much space"


def test_kde_log_get_at():
    ddict = example_data_dict(size=300, keys=["area_um", "deform"])
    ddict["deform"][:20] = .1
    ddict["area_um"][:20] = .5
    ds = dclab.new_dataset(ddict)
    kde_instance = KernelDensityEstimator(ds)
    a = kde_instance.get_at(xax="area_um", yax="deform", yscale="log")
    assert np.all(a[:20] == a[0])


def test_kde_log_get_at_points():
    ddict = example_data_dict(size=300, keys=["area_um", "tilt"])
    ds = dclab.new_dataset(ddict)
    kde_instance = KernelDensityEstimator(ds)
    a = kde_instance.get_at(yscale="log", xax="area_um", yax="tilt")
    b = kde_instance.get_at(yscale="log", xax="area_um", yax="tilt")

    assert np.all(a == b)


def test_kde_log_get_at_invalid():
    ddict = example_data_dict(size=300, keys=["area_um", "deform"])
    ddict["deform"][:20] = .1
    ddict["area_um"][:20] = .5
    ddict["deform"][21] = np.nan
    ddict["deform"][22] = np.inf
    ddict["deform"][23] = -.1
    ds = dclab.new_dataset(ddict)
    kde_instance = KernelDensityEstimator(ds)
    a = kde_instance.get_at(xax="area_um", yax="deform", yscale="log")
    assert np.all(a[:20] == a[0])
    assert np.isnan(a[21])
    assert np.isnan(a[22])
    assert np.isnan(a[23])


def test_kde_get_at_all_nan_inf():
    ddict = example_data_dict()
    ddict["area_um"][:] = [np.nan]
    ddict["area_um"][0] = np.inf

    ds = dclab.new_dataset(ddict)

    kde_instance = KernelDensityEstimator(ds)

    with pytest.warns(KernelDensityEstimationForEmtpyArrayWarning,
                      match="KDE encountered an empty array"):
        sc = kde_instance.get_at(xax="area_um", yax="deform")
    assert np.all(np.isnan(sc))

    ds.config["filtering"]["enable filters"] = False
    with pytest.warns(KernelDensityEstimationForEmtpyArrayWarning,
                      match="KDE encountered an empty array"):
        sc2 = kde_instance.get_at(xax="area_um", yax="deform",
                                  positions=(ds["area_um"], ds["deform"]))
    assert np.all(np.isnan(sc2))


def test_kde_get_at_empty():
    ddict = example_data_dict()
    ddict["area_um"] = np.array([])
    ddict["deform"] = np.array([])

    ds = dclab.new_dataset(ddict)

    kde_instance = KernelDensityEstimator(ds)

    sc = kde_instance.get_at(xax="area_um", yax="deform")
    assert sc.size == 0

    ds.config["filtering"]["enable filters"] = False
    sc2 = kde_instance.get_at(xax="area_um", yax="deform",
                              positions=(ds["area_um"], ds["deform"]))
    assert sc2.size == 0


def test_kde_get_at_positions():
    ddict = example_data_dict()
    ds = dclab.new_dataset(ddict)

    kde_instance = KernelDensityEstimator(ds)

    ds.config["filtering"]["enable filters"] = False
    sc = kde_instance.get_at(xax="area_um", yax="deform")
    sc2 = kde_instance.get_at(xax="area_um", yax="deform",
                              positions=(ds["area_um"], ds["deform"]))
    assert np.all(sc == sc2)


def test_kde_log_get_at_out_of_bounds():
    ddict = example_data_dict(size=300, keys=["area_um", "deform"])
    ds = dclab.new_dataset(ddict)
    kde_instance = KernelDensityEstimator(ds)

    # Define a positions that has values outside the typical data range
    # `area_um` ([0.0, 400]) and `deform` ([0.0, 0.02])
    positions = ([410, 300, 300, 300],
                 [-1, -2, 0.01, 0.015])

    # Get the density at the out-of-bounds position
    a = kde_instance.get_at(xax="area_um", yax="deform",
                            positions=positions, yscale="log")
    assert np.isnan(a[0])
    assert np.isnan(a[1])
    assert np.isfinite(a[2])
    assert np.isfinite(a[3])


def test_kde_get_raster_all_nan_inf():
    ddict = example_data_dict()
    ddict["area_um"][:] = [np.nan]
    ddict["area_um"][0] = np.inf

    ds = dclab.new_dataset(ddict)

    kde_instance = KernelDensityEstimator(ds)
    with pytest.warns(KernelDensityEstimationForEmtpyArrayWarning,
                      match="KDE encountered an empty array"):
        x, y, z = kde_instance.get_raster(xax="area_um", yax="deform")
    assert x.shape == y.shape == z.shape
    assert np.all(np.isnan(z))


def test_kde_get_raster_empty():
    ddict = example_data_dict()
    ddict["area_um"] = np.array([])
    ddict["deform"] = np.array([])

    ds = dclab.new_dataset(ddict)

    kde_instance = KernelDensityEstimator(ds)

    with pytest.warns(KernelDensityEstimationForEmtpyArrayWarning,
                      match="KDE encountered an empty array"):
        x, y, z = kde_instance.get_raster(xax="area_um", yax="deform")
    assert x.shape == y.shape == z.shape
    assert np.all(np.isnan(z))
